from unittest.mock import patch

import numpy as np
import pytest
import statsmodels.api as sm
from pytest import mark
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.linear_model import LassoCV
from sklearn.preprocessing import PolynomialFeatures

import dowhy.datasets
from dowhy import CausalModel
from dowhy.causal_refuters.evalue_sensitivity_analyzer import EValueSensitivityAnalyzer

from .base import TestRefuter


@mark.usefixtures("fixed_seed")
class TestAddUnobservedCommonCauseRefuter(object):
    @mark.parametrize(
        ["error_tolerance", "estimator_method", "effect_strength_on_t", "effect_strength_on_y"],
        [
            (0.01, "backdoor.propensity_score_matching", 0.01, 0.02),
        ],
    )
    def test_refutation_binary_treatment(
        self, error_tolerance, estimator_method, effect_strength_on_t, effect_strength_on_y
    ):
        refuter_tester = TestRefuter(
            error_tolerance,
            estimator_method,
            "add_unobserved_common_cause",
            confounders_effect_on_t="binary_flip",
            confounders_effect_on_y="linear",
            effect_strength_on_t=effect_strength_on_t,
            effect_strength_on_y=effect_strength_on_y,
        )
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "effect_strength_on_t", "effect_strength_on_y"],
        [
            (0.01, "iv.instrumental_variable", 0.01, 0.02),
        ],
    )
    def test_refutation_continuous_treatment(
        self, error_tolerance, estimator_method, effect_strength_on_t, effect_strength_on_y
    ):
        refuter_tester = TestRefuter(
            error_tolerance,
            estimator_method,
            "add_unobserved_common_cause",
            confounders_effect_on_t="linear",
            confounders_effect_on_y="linear",
            effect_strength_on_t=effect_strength_on_t,
            effect_strength_on_y=effect_strength_on_y,
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "effect_strength_on_t", "effect_strength_on_y"],
        [
            (0.01, "iv.instrumental_variable", np.arange(0.01, 0.02, 0.001), np.arange(0.02, 0.03, 0.001)),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_refutation_continuous_treatment_range_both_treatment_outcome(
        self, mock_fig, error_tolerance, estimator_method, effect_strength_on_t, effect_strength_on_y
    ):
        refuter_tester = TestRefuter(
            error_tolerance,
            estimator_method,
            "add_unobserved_common_cause",
            confounders_effect_on_t="linear",
            confounders_effect_on_y="linear",
            effect_strength_on_t=effect_strength_on_t,
            effect_strength_on_y=effect_strength_on_y,
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")
        assert mock_fig.call_count > 0  # we patched figure plotting call to avoid drawing plots during tests

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "effect_strength_on_t", "effect_strength_on_y"],
        [
            (0.01, "iv.instrumental_variable", np.arange(0.01, 0.02, 0.001), 0.02),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_refutation_continuous_treatment_range_treatment(
        self, mock_fig, error_tolerance, estimator_method, effect_strength_on_t, effect_strength_on_y
    ):
        refuter_tester = TestRefuter(
            error_tolerance,
            estimator_method,
            "add_unobserved_common_cause",
            confounders_effect_on_t="linear",
            confounders_effect_on_y="linear",
            effect_strength_on_t=effect_strength_on_t,
            effect_strength_on_y=effect_strength_on_y,
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")
        assert mock_fig.call_count > 0  # we patched figure plotting call to avoid drawing plots during tests

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "effect_strength_on_t", "effect_strength_on_y"],
        [
            (0.01, "iv.instrumental_variable", 0.01, np.arange(0.02, 0.03, 0.001)),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_refutation_continuous_treatment_range_outcome(
        self, mock_fig, error_tolerance, estimator_method, effect_strength_on_t, effect_strength_on_y
    ):
        refuter_tester = TestRefuter(
            error_tolerance,
            estimator_method,
            "add_unobserved_common_cause",
            confounders_effect_on_t="linear",
            confounders_effect_on_y="linear",
            effect_strength_on_t=effect_strength_on_t,
            effect_strength_on_y=effect_strength_on_y,
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")
        assert mock_fig.call_count > 0  # we patched figure plotting call to avoid drawing plots during tests

    @pytest.mark.parametrize(
        ["estimator_method", "effect_fraction_on_treatment", "benchmark_common_causes", "simulation_method"],
        [
            ("backdoor.linear_regression", [1, 2, 3], ["W3"], "linear-partial-R2"),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_linear_sensitivity_with_confounders(
        self, mock_fig, estimator_method, effect_fraction_on_treatment, benchmark_common_causes, simulation_method
    ):
        np.random.seed(100)
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=7,
            num_samples=500,
            num_treatments=1,
            stddev_treatment_noise=10,
            stddev_outcome_noise=1,
        )
        data["df"] = data["df"].drop("W4", axis=1)
        graph_str = 'graph[directed 1node[ id "y" label "y"]node[ id "W0" label "W0"] node[ id "W1" label "W1"] node[ id "W2" label "W2"] node[ id "W3" label "W3"]  node[ id "W5" label "W5"] node[ id "W6" label "W6"]node[ id "v0" label "v0"]edge[source "v0" target "y"]edge[ source "W0" target "v0"] edge[ source "W1" target "v0"] edge[ source "W2" target "v0"] edge[ source "W3" target "v0"] edge[ source "W5" target "v0"] edge[ source "W6" target "v0"]edge[ source "W0" target "y"] edge[ source "W1" target "y"] edge[ source "W2" target "y"] edge[ source "W3" target "y"] edge[ source "W5" target "y"] edge[ source "W6" target "y"]]'
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=graph_str,
            test_significance=None,
        )
        target_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        estimate = model.estimate_effect(target_estimand, method_name=estimator_method)
        ate_estimate = data["ate"]
        refute = model.refute_estimate(
            target_estimand,
            estimate,
            method_name="add_unobserved_common_cause",
            simulation_method=simulation_method,
            benchmark_common_causes=benchmark_common_causes,
            effect_fraction_on_treatment=effect_fraction_on_treatment,
        )
        if refute.confounder_increases_estimate == True:
            bias_adjusted_estimate = refute.benchmarking_results["bias_adjusted_estimate"]
            assert all(
                estimate <= refute.estimate for estimate in bias_adjusted_estimate
            )  # if confounder_increases_estimate is True, adjusted estimate should be lower than original estimate
        else:
            bias_adjusted_estimate = refute.benchmarking_results["bias_adjusted_estimate"]
            assert all(estimate >= refute.estimate for estimate in bias_adjusted_estimate)

        # check if all partial R^2 values are between 0 and 1
        assert all((val >= 0 and val <= 1) for val in refute.benchmarking_results["r2tu_w"])
        assert all((val >= 0 and val <= 1) for val in refute.benchmarking_results["r2yu_tw"])
        assert refute.stats["r2yt_w"] >= 0 and refute.stats["r2yt_w"] <= 1

        assert refute.stats["robustness_value"] >= 0 and refute.stats["robustness_value"] <= 1
        assert mock_fig.call_count > 0  # we patched figure plotting call to avoid drawing plots during tests

    @pytest.mark.parametrize(
        ["estimator_method", "effect_fraction_on_treatment", "benchmark_common_causes", "simulation_method"],
        [
            ("backdoor.linear_regression", [1, 2, 3], ["W3"], "linear-partial-R2"),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_linear_sensitivity_given_strength_of_confounding(
        self, mock_fig, estimator_method, effect_fraction_on_treatment, benchmark_common_causes, simulation_method
    ):
        np.random.seed(100)
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=7,
            num_samples=500,
            num_treatments=1,
            stddev_treatment_noise=10,
            stddev_outcome_noise=1,
        )
        data["df"] = data["df"].drop("W4", axis=1)
        graph_str = 'graph[directed 1node[ id "y" label "y"]node[ id "W0" label "W0"] node[ id "W1" label "W1"] node[ id "W2" label "W2"] node[ id "W3" label "W3"]  node[ id "W5" label "W5"] node[ id "W6" label "W6"]node[ id "v0" label "v0"]edge[source "v0" target "y"]edge[ source "W0" target "v0"] edge[ source "W1" target "v0"] edge[ source "W2" target "v0"] edge[ source "W3" target "v0"] edge[ source "W5" target "v0"] edge[ source "W6" target "v0"]edge[ source "W0" target "y"] edge[ source "W1" target "y"] edge[ source "W2" target "y"] edge[ source "W3" target "y"] edge[ source "W5" target "y"] edge[ source "W6" target "y"]]'
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=graph_str,
            test_significance=None,
        )
        target_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        estimate = model.estimate_effect(target_estimand, method_name=estimator_method)
        ate_estimate = data["ate"]
        refute = model.refute_estimate(
            target_estimand,
            estimate,
            method_name="add_unobserved_common_cause",
            simulation_method=simulation_method,
            benchmark_common_causes=benchmark_common_causes,
            effect_fraction_on_treatment=effect_fraction_on_treatment,
        )

        # We calculate adjusted estimates for two sets of partial R^2 values.
        bias_adjusted_terms = refute.compute_bias_adjusted(r2tu_w=np.array([0.7, 0.2]), r2yu_tw=np.array([0.9, 0.3]))
        estimate1 = bias_adjusted_terms["bias_adjusted_estimate"][
            0
        ]  # adjusted estimate for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        estimate2 = bias_adjusted_terms["bias_adjusted_estimate"][
            1
        ]  # adjusted estimate for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        print(estimate1, estimate2)
        original_estimate = refute.estimate
        # Test if hypothetical confounding by unobserved confounder u1 leads to an adjusted effect that is farther from the original estimate as compared to u2
        assert abs(original_estimate - estimate1) > abs(original_estimate - estimate2)
        assert mock_fig.call_count > 0  # we patched figure plotting call to avoid drawing plots during tests

    @mark.parametrize(
        [
            "estimator_method",
            "effect_fraction_on_treatment",
            "benchmark_common_causes",
            "simulation_method",
            "rvalue_threshold",
        ],
        [
            ("backdoor.linear_regression", [1, 2, 3], ["W3"], "linear-partial-R2", 0.95),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_linear_sensitivity_dataset_without_confounders(
        self,
        mock_fig,
        estimator_method,
        effect_fraction_on_treatment,
        benchmark_common_causes,
        simulation_method,
        rvalue_threshold,
    ):
        np.random.seed(100)
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=7,
            num_samples=500,
            num_treatments=1,
            stddev_treatment_noise=10,
            stddev_outcome_noise=1,
        )
        # Creating a model with no unobserved confounders
        model2 = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=data["gml_graph"],
            test_significance=None,
        )
        target_estimand2 = model2.identify_effect(proceed_when_unidentifiable=True)
        estimate2 = model2.estimate_effect(target_estimand2, method_name=estimator_method)
        ate_estimate = data["ate"]
        refute2 = model2.refute_estimate(
            target_estimand2,
            estimate2,
            method_name="add_unobserved_common_cause",
            simulation_method=simulation_method,
            benchmark_common_causes=benchmark_common_causes,
            effect_fraction_on_treatment=effect_fraction_on_treatment,
        )

        if refute2.confounder_increases_estimate == True:
            bias_adjusted_estimate = refute2.benchmarking_results["bias_adjusted_estimate"]
            assert all(
                estimate <= refute2.estimate for estimate in bias_adjusted_estimate
            )  # if confounder_increases_estimate is True, adjusted estimate should be lower than original estimate
        else:
            bias_adjusted_estimate = refute2.benchmarking_results["bias_adjusted_estimate"]
            assert all(estimate >= refute2.estimate for estimate in bias_adjusted_estimate)

        # check if all partial R^2 values are between 0 and 1
        assert all((val >= 0 and val <= 1) for val in refute2.benchmarking_results["r2tu_w"])
        assert all((val >= 0 and val <= 1) for val in refute2.benchmarking_results["r2yu_tw"])
        assert refute2.stats["r2yt_w"] >= 0 and refute2.stats["r2yt_w"] <= 1

        print(refute2.stats["robustness_value"])
        # for a dataset with no confounders, the robustness value should be higher than a given threshold (0.95 in our case)
        assert refute2.stats["robustness_value"] >= rvalue_threshold and refute2.stats["robustness_value"] <= 1
        assert mock_fig.call_count > 0  # we patched figure plotting call to avoid drawing plots during tests

    @pytest.mark.parametrize(
        ["estimator_method", "effect_fraction_on_treatment", "benchmark_common_causes", "simulation_method"],
        [
            ("backdoor.econml.dml.KernelDML", 2, ["W3"], "non-parametric-partial-R2"),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_non_parametric_sensitivity_given_strength_of_confounding(
        self, mock_fig, estimator_method, effect_fraction_on_treatment, benchmark_common_causes, simulation_method
    ):
        np.random.seed(100)
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=7,
            num_samples=500,
            num_treatments=1,
            stddev_treatment_noise=5,
            stddev_outcome_noise=5,
        )
        data["df"] = data["df"].drop("W4", axis=1)
        graph_str = 'graph[directed 1node[ id "y" label "y"]node[ id "W0" label "W0"] node[ id "W1" label "W1"] node[ id "W2" label "W2"] node[ id "W3" label "W3"]  node[ id "W5" label "W5"] node[ id "W6" label "W6"]node[ id "v0" label "v0"]edge[source "v0" target "y"]edge[ source "W0" target "v0"] edge[ source "W1" target "v0"] edge[ source "W2" target "v0"] edge[ source "W3" target "v0"] edge[ source "W5" target "v0"] edge[ source "W6" target "v0"]edge[ source "W0" target "y"] edge[ source "W1" target "y"] edge[ source "W2" target "y"] edge[ source "W3" target "y"] edge[ source "W5" target "y"] edge[ source "W6" target "y"]]'
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=graph_str,
            test_significance=None,
        )
        target_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        # Non Parametric estimator
        estimate = model.estimate_effect(
            target_estimand,
            method_name=estimator_method,
            method_params={
                "init_params": {
                    "model_y": GradientBoostingRegressor(),
                    "model_t": GradientBoostingRegressor(),
                },
                "fit_params": {
                    "cache_values": True,
                },
            },
        )
        ate_estimate = data["ate"]
        refute = model.refute_estimate(
            target_estimand,
            estimate,
            method_name="add_unobserved_common_cause",
            simulation_method=simulation_method,
            benchmark_common_causes=benchmark_common_causes,
            effect_fraction_on_treatment=effect_fraction_on_treatment,
        )

        assert refute.r2yu_tw >= 0 and refute.r2yu_tw <= 1
        assert refute.r2tu_w >= 0 and refute.r2tu_w <= 1

        # We calculate adjusted estimates for two sets of partial R^2 values.
        benchmarking_results_u1 = refute.perform_benchmarking(r2yu_tw=0.9, r2tu_w=0.7, significance_level=0.05)
        benchmarking_results_u2 = refute.perform_benchmarking(r2yu_tw=0.3, r2tu_w=0.2, significance_level=0.05)
        # adjusted lower ate bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        lower_ate_bound_u1 = benchmarking_results_u1["lower_ate_bound"]
        # adjusted lower ate bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        lower_ate_bound_u2 = benchmarking_results_u2["lower_ate_bound"]

        # adjusted upper ate bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        upper_ate_bound_u1 = benchmarking_results_u1["upper_ate_bound"]
        # adjusted upper ate bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        upper_ate_bound_u2 = benchmarking_results_u2["upper_ate_bound"]

        # adjusted lower confidence bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        lower_confidence_bound_u1 = benchmarking_results_u1["lower_confidence_bound"]
        # adjusted lower confidence bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        lower_confidence_bound_u2 = benchmarking_results_u2["lower_confidence_bound"]

        # adjusted upper confidence bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        upper_confidence_bound_u1 = benchmarking_results_u1["upper_confidence_bound"]
        # adjusted upper confidence bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        upper_confidence_bound_u2 = benchmarking_results_u2["upper_confidence_bound"]

        original_estimate = refute.theta_s
        # Test if hypothetical confounding by unobserved confounder u1 leads to an adjusted effect that is farther from the original estimate as compared to u2
        assert abs(original_estimate - lower_ate_bound_u1) > abs(original_estimate - lower_ate_bound_u2)
        assert abs(original_estimate - upper_ate_bound_u1) > abs(original_estimate - upper_ate_bound_u2)
        assert abs(original_estimate - lower_confidence_bound_u1) > abs(original_estimate - lower_confidence_bound_u2)
        assert abs(original_estimate - upper_confidence_bound_u1) > abs(original_estimate - upper_confidence_bound_u2)
        # we patched figure plotting call to avoid drawing plots during tests
        assert mock_fig.call_count > 0

    @pytest.mark.parametrize(
        ["estimator_method", "effect_fraction_on_outcome", "benchmark_common_causes", "simulation_method"],
        [
            ("backdoor.econml.dml.dml.LinearDML", 2, ["W3"], "non-parametric-partial-R2"),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_partially_linear_sensitivity_given_strength_of_confounding(
        self, mock_fig, estimator_method, effect_fraction_on_outcome, benchmark_common_causes, simulation_method
    ):
        np.random.seed(100)
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=7,
            num_samples=500,
            num_treatments=1,
            stddev_treatment_noise=5,
            stddev_outcome_noise=5,
        )
        data["df"] = data["df"].drop("W4", axis=1)
        graph_str = 'graph[directed 1node[ id "y" label "y"]node[ id "W0" label "W0"] node[ id "W1" label "W1"] node[ id "W2" label "W2"] node[ id "W3" label "W3"]  node[ id "W5" label "W5"] node[ id "W6" label "W6"]node[ id "v0" label "v0"]edge[source "v0" target "y"]edge[ source "W0" target "v0"] edge[ source "W1" target "v0"] edge[ source "W2" target "v0"] edge[ source "W3" target "v0"] edge[ source "W5" target "v0"] edge[ source "W6" target "v0"]edge[ source "W0" target "y"] edge[ source "W1" target "y"] edge[ source "W2" target "y"] edge[ source "W3" target "y"] edge[ source "W5" target "y"] edge[ source "W6" target "y"]]'
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=graph_str,
            test_significance=None,
        )
        target_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        estimate = model.estimate_effect(
            target_estimand,
            method_name=estimator_method,
            method_params={
                "init_params": {
                    "model_y": GradientBoostingRegressor(),
                    "model_t": GradientBoostingRegressor(),
                    "linear_first_stages": False,
                },
                "fit_params": {
                    "cache_values": True,
                },
            },
        )
        ate_estimate = data["ate"]
        refute = model.refute_estimate(
            target_estimand,
            estimate,
            method_name="add_unobserved_common_cause",
            simulation_method=simulation_method,
            benchmark_common_causes=benchmark_common_causes,
            effect_fraction_on_outcome=effect_fraction_on_outcome,
        )

        assert refute.r2yu_tw >= 0 and refute.r2yu_tw <= 1
        assert refute.r2tu_w >= 0 and refute.r2tu_w <= 1

        # We calculate adjusted estimates for two sets of partial R^2 values.
        benchmarking_results_u1 = refute.perform_benchmarking(r2yu_tw=0.9, r2tu_w=0.7, significance_level=0.05)
        benchmarking_results_u2 = refute.perform_benchmarking(r2yu_tw=0.3, r2tu_w=0.2, significance_level=0.05)
        # adjusted lower ate bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        lower_ate_bound_u1 = benchmarking_results_u1["lower_ate_bound"]
        # adjusted lower ate bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        lower_ate_bound_u2 = benchmarking_results_u2["lower_ate_bound"]

        # adjusted upper ate bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        upper_ate_bound_u1 = benchmarking_results_u1["upper_ate_bound"]
        # adjusted upper ate bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        upper_ate_bound_u2 = benchmarking_results_u2["upper_ate_bound"]

        # adjusted lower confidence bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        lower_confidence_bound_u1 = benchmarking_results_u1["lower_confidence_bound"]
        # adjusted lower confidence bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        lower_confidence_bound_u2 = benchmarking_results_u2["lower_confidence_bound"]

        # adjusted upper confidence bound for confounder u1 where r2tu_w = 0.7 and r2yu_tw = 0.9
        upper_confidence_bound_u1 = benchmarking_results_u1["upper_confidence_bound"]
        # adjusted upper confidence bound for confounder u2 where r2tu_w = 0.2 and r2yu_tw = 0.3
        upper_confidence_bound_u2 = benchmarking_results_u2["upper_confidence_bound"]

        original_estimate = refute.theta_s
        # Test if hypothetical confounding by unobserved confounder u1 leads to an adjusted effect that is farther from the original estimate as compared to u2
        assert abs(original_estimate - lower_ate_bound_u1) > abs(original_estimate - lower_ate_bound_u2)
        assert abs(original_estimate - upper_ate_bound_u1) > abs(original_estimate - upper_ate_bound_u2)
        assert abs(original_estimate - lower_confidence_bound_u1) > abs(original_estimate - lower_confidence_bound_u2)
        assert abs(original_estimate - upper_confidence_bound_u1) > abs(original_estimate - upper_confidence_bound_u2)
        # we patched figure plotting call to avoid drawing plots during tests
        assert mock_fig.call_count > 0

    def test_evalue(self):
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=5,
            num_samples=1000,
            treatment_is_binary=True,
        )
        model = CausalModel(
            data=data["df"], treatment=data["treatment_name"], outcome=data["outcome_name"], graph=data["gml_graph"]
        )
        identified_estimand = model.identify_effect()
        estimate = model.estimate_effect(identified_estimand, method_name="backdoor.linear_regression")
        outcome_vals = data["df"][data["outcome_name"]]
        analyzer = EValueSensitivityAnalyzer(
            estimate, identified_estimand, data["df"], data["treatment_name"], data["outcome_name"]
        )

        # comparing test examples from R E-Value package
        stats = analyzer._evalue_OLS(0.293, 0.196, 17.405)
        assert stats["converted_estimate"].round(3) == 1.015
        assert stats["converted_lower_ci"].round(3) == 0.995
        assert stats["converted_upper_ci"].round(3) == 1.036
        assert stats["evalue_estimate"].round(3) == 1.141
        assert stats["evalue_lower_ci"].round(3) == 1.000
        assert stats["evalue_upper_ci"] is None

        stats = analyzer._evalue_MD(0.5, 0.25)
        assert stats["converted_estimate"].round(3) == 1.576
        assert stats["converted_lower_ci"].round(3) == 1.010
        assert stats["converted_upper_ci"].round(3) == 2.460
        assert stats["evalue_estimate"].round(3) == 2.529
        assert stats["evalue_lower_ci"].round(3) == 1.111
        assert stats["evalue_upper_ci"] is None

        stats = analyzer._evalue_OR(0.86, 0.75, 0.99, rare=False)
        assert stats["converted_estimate"].round(3) == 0.927
        assert stats["converted_lower_ci"].round(3) == 0.866
        assert stats["converted_upper_ci"].round(3) == 0.995
        assert stats["evalue_estimate"].round(3) == 1.369
        assert stats["evalue_lower_ci"] == None
        assert stats["evalue_upper_ci"].round(3) == 1.076

        stats = analyzer._evalue_RR(0.8, 0.71, 0.91)
        assert stats["converted_estimate"] == 0.8
        assert stats["converted_lower_ci"] == 0.71
        assert stats["converted_upper_ci"] == 0.91
        assert stats["evalue_estimate"].round(3) == 1.809
        assert stats["evalue_lower_ci"] == None
        assert stats["evalue_upper_ci"].round(3) == 1.429

        # check implementation of Observed Covariate E-value against R package
        assert analyzer._observed_covariate_e_value(2, 4).round(3) == 3.414
        assert analyzer._observed_covariate_e_value(4, 2).round(3) == 3.414
        assert analyzer._observed_covariate_e_value(0.8, 0.9).round(3) == 1.5
        assert analyzer._observed_covariate_e_value(0.9, 0.8).round(3) == 1.5

    @pytest.mark.parametrize(
        "estimator_method",
        [
            ("backdoor.linear_regression"),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_evalue_linear_regression(self, mock_fig, estimator_method):
        data = dowhy.datasets.linear_dataset(
            beta=10, num_common_causes=5, num_samples=1000, treatment_is_binary=True, stddev_outcome_noise=5
        )
        model = CausalModel(
            data=data["df"], treatment=data["treatment_name"], outcome=data["outcome_name"], graph=data["gml_graph"]
        )
        identified_estimand = model.identify_effect()
        estimate = model.estimate_effect(identified_estimand, method_name=estimator_method)
        refute = model.refute_estimate(
            identified_estimand, estimate, method_name="add_unobserved_common_cause", simulation_method="e-value"
        )

        assert refute.stats["evalue_upper_ci"] is None
        assert refute.stats["evalue_lower_ci"] < refute.stats["evalue_estimate"]
        assert mock_fig.call_count > 0

    @pytest.mark.parametrize(
        "estimator_method",
        [
            ("backdoor.generalized_linear_model"),
        ],
    )
    @patch("matplotlib.pyplot.figure")
    def test_evalue_logistic_regression(self, mock_fig, estimator_method):
        data = dowhy.datasets.linear_dataset(
            beta=10,
            outcome_is_binary=True,
            num_common_causes=5,
            num_samples=1000,
            treatment_is_binary=True,
            stddev_outcome_noise=5,
        )
        model = CausalModel(
            data=data["df"], treatment=data["treatment_name"], outcome=data["outcome_name"], graph=data["gml_graph"]
        )
        identified_estimand = model.identify_effect()
        estimate = model.estimate_effect(
            identified_estimand, method_name=estimator_method, method_params={"glm_family": sm.families.Binomial()}
        )
        refute = model.refute_estimate(
            identified_estimand, estimate, method_name="add_unobserved_common_cause", simulation_method="e-value"
        )

        assert refute.stats["evalue_upper_ci"] is None
        assert refute.stats["evalue_lower_ci"] < refute.stats["evalue_estimate"]
        assert mock_fig.call_count > 0



show_progress_bars = True
default_n_jobs = -1


def enable_progress_bars():
    global show_progress_bars
    show_progress_bars = True


def disable_progress_bars():
    global show_progress_bars
    show_progress_bars = False


def set_default_n_jobs(n_jobs: int) -> None:
    global default_n_jobs
    default_n_jobs = n_jobs



import os
import subprocess
import tempfile

import nbformat
import pytest
from pytest import mark

NOTEBOOKS_PATH = "docs/source/example_notebooks/"
notebooks_list = [f.name for f in os.scandir(NOTEBOOKS_PATH) if f.name.endswith(".ipynb")]
advanced_notebooks = [
    # requires stdin input for identify in weighting sampler
    "do_sampler_demo.ipynb",
    # requires Rpy2 for lalonde
    "dowhy_refutation_testing.ipynb",
    "dowhy_lalonde_example.ipynb",
    "lalonde_pandas_api.ipynb",
    # will be removed
    "dowhy_optimize_backdoor_example.ipynb",
    # applied notebook, not necessary to test each time
    "dowhy_ranking_methods.ipynb",
    # needs xgboost too
    "DoWhy-The Causal Story Behind Hotel Booking Cancellations.ipynb",
    # Slow Notebooks
    "dowhy-conditional-treatment-effects.ipynb",
    "dowhy_refuter_notebook.ipynb",
    "dowhy_twins_example.ipynb",
]

ignore_notebooks = [
    # requires Rpy2 for causal discovery
    # daily tests of dowhy_causal_discovery_example.ipynb are failing due to cdt/rpy2 config.
    # comment out, since we are switching causal discovery implementations
    "dowhy_causal_discovery_example.ipynb"
]

# Adding the dowhy root folder to the python path so that jupyter notebooks
# can import dowhy
if "PYTHONPATH" not in os.environ:
    os.environ["PYTHONPATH"] = os.getcwd()
elif os.getcwd() not in os.environ["PYTHONPATH"].split(os.pathsep):
    os.environ["PYTHONPATH"] = os.environ["PYTHONPATH"] + os.pathsep + os.getcwd()


def _notebook_run(filepath):
    """Execute a notebook via nbconvert and collect output.
    :returns (parsed nb object, execution errors)

    Source of this function: http://www.christianmoscardi.com/blog/2016/01/20/jupyter-testing.html
    """
    with tempfile.NamedTemporaryFile(suffix=".ipynb") as fout:
        args = [
            "jupyter",
            "nbconvert",
            "--to",
            "notebook",
            "--execute",
            #          "--ExecutePreprocessor.timeout=600",
            "-y",
            "--no-prompt",
            "--output",
            fout.name,
            filepath,
        ]
        subprocess.check_call(args)

        fout.seek(0)
        nb = nbformat.read(fout, nbformat.current_nbformat)

    errors = [
        output for cell in nb.cells if "outputs" in cell for output in cell["outputs"] if output.output_type == "error"
    ]

    return nb, errors


"""
def test_getstarted_notebook():
    nb, errors = _notebook_run(NOTEBOOKS_PATH+ "dowhy_simple_example.ipynb")
    assert errors == []

def test_confounder_notebook():
    nb, errors = _notebook_run(NOTEBOOKS_PATH+"dowhy_confounder_example.ipynb")
    assert errors = []
"""
parameter_list = []
for nb in notebooks_list:
    if nb in ignore_notebooks:
        continue
    elif nb in advanced_notebooks:
        param = pytest.param(nb, marks=[mark.advanced, mark.notebook], id=nb)
    else:
        param = pytest.param(nb, marks=[mark.notebook], id=nb)
    parameter_list.append(param)


@mark.parametrize("notebook_filename", parameter_list)
def test_notebook(notebook_filename):
    nb, errors = _notebook_run(NOTEBOOKS_PATH + notebook_filename)
    assert errors == []



from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.propensity_score_estimator import PropensityScoreEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class PropensityScoreWeightingEstimator(PropensityScoreEstimator):
    """Estimate effect of treatment by weighing the data by
    inverse probability of occurrence.

    Straightforward application of the back-door criterion.

    Supports additional parameters as listed below.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        min_ps_score: float = 0.05,
        max_ps_score: float = 0.95,
        weighting_scheme: str = "ips_weight",
        propensity_score_model: Optional[Any] = None,
        propensity_score_column: str = "propensity_score",
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param min_ps_score: Lower bound used to clip the propensity score.
            Default=0.05
        :param max_ps_score: Upper bound used to clip the propensity score.
            Default=0.95
        :param weighting_scheme: Weighting method to use. Can be inverse
            propensity score ("ips_weight", default), stabilized IPS score
            ("ips_stabilized_weight"), or normalized IPS score
            ("ips_normalized_weight").
        :param propensity_score_model: The model used to compute propensity
            score. Can be any classification model that supports fit() and
            predict_proba() methods. If None, use LogisticRegression model as
            the default. Default=None
        :param propensity_score_column: Column name that stores the
            propensity score. Default='propensity_score'
        :param kwargs: (optional) Additional estimator-specific parameters
        """

        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            propensity_score_model=propensity_score_model,
            propensity_score_column=propensity_score_column,
            min_ps_score=min_ps_score,
            max_ps_score=max_ps_score,
            weighting_scheme=weighting_scheme,
            **kwargs,
        )

        self.logger.info("INFO: Using Propensity Score Weighting Estimator")

        # Setting method specific parameters
        self.weighting_scheme = weighting_scheme
        self.min_ps_score = min_ps_score
        self.max_ps_score = max_ps_score

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        super().fit(data, treatment_name, outcome_name, effect_modifier_names=effect_modifier_names)

        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        return self

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        if data is None:
            data = self._data
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value
        if self.propensity_score_column not in data:
            self.estimate_propensity_score_column(data)

        # trim propensity score weights
        data[self.propensity_score_column] = np.minimum(self.max_ps_score, data[self.propensity_score_column])
        data[self.propensity_score_column] = np.maximum(self.min_ps_score, data[self.propensity_score_column])

        # ips ==> (isTreated(y)/ps(y)) + ((1-isTreated(y))/(1-ps(y)))
        # nips ==> ips / (sum of ips over all units)
        # icps ==> ps(y)/(1-ps(y)) / (sum of (ps(y)/(1-ps(y))) over all control units)
        # itps ==> ps(y)/(1-ps(y)) / (sum of (ps(y)/(1-ps(y))) over all treatment units)
        ipst_sum = sum(data[self._treatment_name[0]] / data[self.propensity_score_column])
        ipsc_sum = sum((1 - data[self._treatment_name[0]]) / (1 - data[self.propensity_score_column]))
        num_units = len(data[self._treatment_name[0]])
        num_treatment_units = sum(data[self._treatment_name[0]])
        num_control_units = num_units - num_treatment_units

        # Vanilla IPS estimator
        data["ips_weight"] = data[self._treatment_name[0]] / data[self.propensity_score_column] + (
            1 - data[self._treatment_name[0]]
        ) / (1 - data[self.propensity_score_column])
        data["tips_weight"] = data[self._treatment_name[0]] + (1 - data[self._treatment_name[0]]) * data[
            self.propensity_score_column
        ] / (1 - data[self.propensity_score_column])
        data["cips_weight"] = data[self._treatment_name[0]] * (1 - data[self.propensity_score_column]) / data[
            self.propensity_score_column
        ] + (1 - data[self._treatment_name[0]])

        # The Hajek estimator (or the self-normalized estimator)
        data["ips_normalized_weight"] = (
            data[self._treatment_name[0]] / data[self.propensity_score_column] / ipst_sum
            + (1 - data[self._treatment_name[0]]) / (1 - data[self.propensity_score_column]) / ipsc_sum
        )
        ipst_for_att_sum = sum(data[self._treatment_name[0]])
        ipsc_for_att_sum = sum(
            (1 - data[self._treatment_name[0]])
            / (1 - data[self.propensity_score_column])
            * data[self.propensity_score_column]
        )
        data["tips_normalized_weight"] = (
            data[self._treatment_name[0]] / ipst_for_att_sum
            + (1 - data[self._treatment_name[0]])
            * data[self.propensity_score_column]
            / (1 - data[self.propensity_score_column])
            / ipsc_for_att_sum
        )
        ipst_for_atc_sum = sum(
            data[self._treatment_name[0]]
            / data[self.propensity_score_column]
            * (1 - data[self.propensity_score_column])
        )
        ipsc_for_atc_sum = sum((1 - data[self._treatment_name[0]]))
        data["cips_normalized_weight"] = (
            data[self._treatment_name[0]]
            * (1 - data[self.propensity_score_column])
            / data[self.propensity_score_column]
            / ipst_for_atc_sum
            + (1 - data[self._treatment_name[0]]) / ipsc_for_atc_sum
        )

        # Stabilized weights (from Robins, Hernan, Brumback (2000))
        # Paper: Marginal Structural Models and Causal Inference in Epidemiology
        p_treatment = sum(data[self._treatment_name[0]]) / num_units
        data["ips_stabilized_weight"] = data[self._treatment_name[0]] / data[
            self.propensity_score_column
        ] * p_treatment + (1 - data[self._treatment_name[0]]) / (1 - data[self.propensity_score_column]) * (
            1 - p_treatment
        )
        data["tips_stabilized_weight"] = data[self._treatment_name[0]] * p_treatment + (
            1 - data[self._treatment_name[0]]
        ) * data[self.propensity_score_column] / (1 - data[self.propensity_score_column]) * (1 - p_treatment)
        data["cips_stabilized_weight"] = data[self._treatment_name[0]] * (
            1 - data[self.propensity_score_column]
        ) / data[self.propensity_score_column] * p_treatment + (1 - data[self._treatment_name[0]]) * (1 - p_treatment)

        if isinstance(target_units, pd.DataFrame) or target_units == "ate":
            weighting_scheme_name = self.weighting_scheme
        elif target_units == "att":
            weighting_scheme_name = "t" + self.weighting_scheme
        elif target_units == "atc":
            weighting_scheme_name = "c" + self.weighting_scheme
        else:
            raise ValueError(f"Target units value {target_units} not supported")

        # Calculating the effect
        data["d_y"] = data[weighting_scheme_name] * data[self._treatment_name[0]] * data[self._outcome_name]
        data["dbar_y"] = data[weighting_scheme_name] * (1 - data[self._treatment_name[0]]) * data[self._outcome_name]
        sum_dy_weights = np.sum(data[self._treatment_name[0]] * data[weighting_scheme_name])
        sum_dbary_weights = np.sum((1 - data[self._treatment_name[0]]) * data[weighting_scheme_name])
        # Subtracting the weighted means
        est = data["d_y"].sum() / sum_dy_weights - data["dbar_y"].sum() / sum_dbary_weights

        # TODO - how can we add additional information into the returned estimate?
        estimate = CausalEstimate(
            estimate=est,
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
            propensity_scores=data[self.propensity_score_column],
        )

        estimate.add_estimator(self)
        return estimate

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ",".join(estimand.outcome_variable) + "~"
        # TODO -- fix: we are actually conditioning on positive treatment (d=1)
        var_list = estimand.treatment_variable + estimand.get_backdoor_variables()
        expr += "+".join(var_list)
        return expr



import random

import networkx as nx
import numpy as np
import pandas as pd
import pytest
from flaky import flaky
from pytest import approx
from scipy import stats

from dowhy.gcm import (
    AdditiveNoiseModel,
    ClassifierFCM,
    ProbabilisticCausalModel,
    ScipyDistribution,
    arrow_strength,
    fit,
)
from dowhy.gcm.divergence import estimate_kl_divergence_continuous
from dowhy.gcm.influence import arrow_strength_of_model
from dowhy.gcm.ml import create_linear_regressor, create_logistic_regression_classifier


@pytest.fixture
def preserve_random_generator_state():
    numpy_state = np.random.get_state()
    random_state = random.getstate()
    yield
    np.random.set_state(numpy_state)
    random.setstate(random_state)


@flaky(max_runs=5)
def test_given_kl_divergence_attribution_func_when_estimate_arrow_strength_then_returns_expected_results():
    causal_strengths = arrow_strength(
        _create_causal_model(), "X2", difference_estimation_func=estimate_kl_divergence_continuous
    )
    assert causal_strengths[("X0", "X2")] == approx(2.76, abs=0.4)
    assert causal_strengths[("X1", "X2")] == approx(1.6, abs=0.4)


@flaky(max_runs=5)
def test_given_continuous_data_with_default_attribution_func_when_estimate_arrow_strength_then_returns_expected_results():
    causal_strengths = arrow_strength(_create_causal_model(), "X2")

    # By default, the strength is measure with respect to the variance.
    assert causal_strengths[("X0", "X2")] == approx(9, abs=0.5)
    assert causal_strengths[("X1", "X2")] == approx(1, abs=0.2)


@flaky(max_runs=3)
def test_given_gcm_with_misspecified_mechanism_when_evaluate_arrow_strength_with__observational_data_then_gives_expected_results():
    causal_model = ProbabilisticCausalModel(nx.DiGraph([("X1", "X2"), ("X0", "X2")]))
    # Here, we misspecified the mechanism on purpose by setting scale to 1 instead of 2.
    causal_model.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X1", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    X0 = np.random.normal(0, 2, 2000)  # The standard deviation in the data is actually 2.
    X1 = np.random.normal(0, 1, 2000)

    test_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X0 + X1 + np.random.normal(0, 0.2, X0.shape[0])})
    fit(causal_model, test_data)

    # If we provide the observational data here, we can mitigate the misspecification of the causal mechanism.
    causal_strengths = arrow_strength(
        causal_model, "X2", parent_samples=test_data, difference_estimation_func=lambda x, y: np.var(y) - np.var(x)
    )
    assert causal_strengths[("X0", "X2")] == approx(4, abs=0.5)
    assert causal_strengths[("X1", "X2")] == approx(1, abs=0.1)


@flaky(max_runs=5)
def test_given_categorical_target_node_when_estimate_arrow_strength_of_model_classifier_then_returns_expected_result():
    X = np.random.random((1000, 5))
    Y = []

    for n in X:
        if (n[0] + n[1] + np.random.random(1)) > 1.5:
            Y.append(0)
        else:
            Y.append(1)

    classifier_sem = ClassifierFCM(create_logistic_regression_classifier())
    classifier_sem.fit(X, np.array(Y).astype(str))

    assert arrow_strength_of_model(classifier_sem, X) == approx(np.array([0.3, 0.3, 0, 0, 0]), abs=0.1)


def _create_causal_model():
    causal_model = ProbabilisticCausalModel(nx.DiGraph([("X1", "X2"), ("X0", "X2")]))
    causal_model.set_causal_mechanism("X1", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    X0 = np.random.normal(0, 1, 1000)
    X1 = np.random.normal(0, 1, 1000)

    test_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": 3 * X0 + X1 + np.random.normal(0, 0.2, X0.shape[0])})
    fit(causal_model, test_data)

    return causal_model



import random

import numpy as np
import pytest
from flaky import flaky

from dowhy.gcm.independence_test import approx_kernel_based, kernel_based


@flaky(max_runs=5)
def test_given_continuous_conditionally_independent_data_when_perform_kernel_based_test_then_not_reject():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert kernel_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_given_continuous_conditionally_dependent_data_when_perform_kernel_based_test_then_reject():
    z = np.random.randn(1000, 1)
    w = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert 0.05 > kernel_based(x, y, w)


@flaky(max_runs=5)
def test_given_categorical_conditionally_independent_data_when_perform_kernel_based_test_then_not_reject():
    x, y, z = _generate_categorical_data()

    assert kernel_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_given_categorical_conditionally_dependent_data_when_perform_kernel_based_test_then_reject():
    x, y, z = _generate_categorical_data()

    assert kernel_based(x, z, y) < 0.05


@flaky(max_runs=2)
def test_given_random_seed_when_perform_conditional_kernel_based_test_then_return_deterministic_result(
    _preserve_random_generator_state,
):
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert kernel_based(
        x, z, y, bootstrap_num_samples_per_run=5, bootstrap_num_runs=2, p_value_adjust_func=np.mean
    ) != kernel_based(x, z, y, bootstrap_num_samples_per_run=5, bootstrap_num_runs=2, p_value_adjust_func=np.mean)

    np.random.seed(0)
    result_1 = kernel_based(x, z, y, bootstrap_num_samples_per_run=5, bootstrap_num_runs=2, p_value_adjust_func=np.mean)
    np.random.seed(0)
    result_2 = kernel_based(x, z, y, bootstrap_num_samples_per_run=5, bootstrap_num_runs=2, p_value_adjust_func=np.mean)

    assert result_1 == result_2


@flaky(max_runs=5)
def test_given_continuous_independent_data_when_perform_kernel_based_test_then_not_reject():
    x = np.random.randn(1000, 1)
    y = np.exp(np.random.rand(1000, 1))

    assert kernel_based(x, y) > 0.05


@flaky(max_runs=5)
def test_given_continuous_dependent_data_when_perform_kernel_based_test_then_reject():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert kernel_based(x, y) < 0.05


@flaky(max_runs=5)
def test_given_categorical_independent_data_when_perform_kernel_based_test_then_not_reject():
    x = np.random.normal(0, 1, 1000)
    y = (np.random.choice(2, 1000) == 1).astype(str)

    assert kernel_based(x, y) > 0.05


@flaky(max_runs=5)
def test_given_categorical_dependent_data_when_perform_kernel_based_test_then_reject():
    x = np.random.normal(0, 1, 1000)
    y = []

    for v in x:
        if v > 0:
            y.append(0)
        else:
            y.append(1)
    y = np.array(y).astype(str)

    assert kernel_based(x, y) < 0.05


@flaky(max_runs=2)
def test_given_random_seed_when_perform_pairwise_kernel_based_test_then_return_deterministic_result(
    _preserve_random_generator_state,
):
    x = np.random.randn(1000, 1)
    y = x + np.random.randn(1000, 1)

    assert kernel_based(
        x, y, bootstrap_num_samples_per_run=10, bootstrap_num_runs=2, p_value_adjust_func=np.mean
    ) != kernel_based(x, y, bootstrap_num_samples_per_run=10, bootstrap_num_runs=2, p_value_adjust_func=np.mean)

    np.random.seed(0)
    result_1 = kernel_based(x, y, bootstrap_num_samples_per_run=10, bootstrap_num_runs=2, p_value_adjust_func=np.mean)
    np.random.seed(0)
    result_2 = kernel_based(x, y, bootstrap_num_samples_per_run=10, bootstrap_num_runs=2, p_value_adjust_func=np.mean)

    assert result_1 == result_2


def test_given_constant_inputs_when_perform_kernel_based_test_then_returns_non_nan_value():
    assert kernel_based(np.random.normal(0, 1, (1000, 2)), np.array([5] * 1000)) != np.nan


@flaky(max_runs=5)
def test_given_continuous_conditionally_independent_data_when_perform_approx_kernel_based_test_then_not_reject():
    z = np.random.randn(5000, 1)
    x = np.exp(z + np.random.rand(5000, 1))
    y = np.exp(z + np.random.rand(5000, 1))

    assert (
        approx_kernel_based(x, y, z, num_random_features_X=10, num_random_features_Y=10, num_random_features_Z=10)
        > 0.05
    )


@flaky(max_runs=5)
def test_given_continuous_conditionally_dependent_data_when_perform_approx_kernel_based_test_then_reject():
    z = np.random.randn(1000, 1)
    w = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert 0.05 > approx_kernel_based(x, y, w)


@flaky(max_runs=5)
def test_given_categorical_conditionally_independent_data_when_perform_approx_kernel_based_test_then_not_reject():
    x, y, z = _generate_categorical_data()

    assert approx_kernel_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_given_categorical_conditionally_dependent_data_when_perform_approx_kernel_based_test_then_reject():
    x, y, z = _generate_categorical_data()

    assert approx_kernel_based(x, z, y) < 0.05


@flaky(max_runs=2)
def test_given_random_seed_when_perform_conditional_approx_kernel_based_test_then_return_deterministic_result(
    _preserve_random_generator_state,
):
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert approx_kernel_based(
        x,
        z,
        y,
        num_random_features_X=1,
        num_random_features_Y=1,
        num_random_features_Z=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    ) != approx_kernel_based(
        x,
        z,
        y,
        num_random_features_X=1,
        num_random_features_Y=1,
        num_random_features_Z=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    )

    np.random.seed(0)
    result_1 = approx_kernel_based(
        x,
        z,
        y,
        num_random_features_X=1,
        num_random_features_Y=1,
        num_random_features_Z=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    )
    np.random.seed(0)
    result_2 = approx_kernel_based(
        x,
        z,
        y,
        num_random_features_X=1,
        num_random_features_Y=1,
        num_random_features_Z=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    )

    assert result_1 == result_2


@flaky(max_runs=5)
def test_given_continuous_independent_data_when_perform_approx_kernel_based_test_then_not_reject():
    x = np.random.randn(1000, 1)
    y = np.exp(np.random.rand(1000, 1))
    assert approx_kernel_based(x, y) > 0.05


@flaky(max_runs=5)
def test_given_continuous_dependent_data_when_perform_approx_kernel_based_test_then_reject():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert approx_kernel_based(x, y) < 0.05


@flaky(max_runs=5)
def test_given_categorical_independent_data_when_perform_approx_kernel_based_test_then_not_reject():
    x = np.random.normal(0, 1, 1000)
    y = np.random.choice(2, 1000).astype(str)
    y[y == "0"] = "Class 1"
    y[y == "1"] = "Class 2"

    assert approx_kernel_based(x, y) > 0.05


@flaky(max_runs=5)
def test_given_categorical_dependent_data_when_perform_approx_kernel_based_test_then_reject():
    x = np.random.normal(0, 1, 1000)
    y = []

    for v in x:
        if v > 0:
            y.append("Class 1")
        else:
            y.append("Class 2")
    y = np.array(y).astype(str)

    assert approx_kernel_based(x, y) < 0.05


@flaky(max_runs=2)
def test_given_random_seed_when_perform_pairwise_approx_kernel_based_test_then_return_deterministic_result(
    _preserve_random_generator_state,
):
    w = np.random.randn(1000, 1)
    x = w + np.random.rand(1000, 1)

    assert approx_kernel_based(
        x,
        w,
        num_random_features_X=1,
        num_random_features_Y=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    ) != approx_kernel_based(
        x,
        w,
        num_random_features_X=1,
        num_random_features_Y=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    )

    np.random.seed(0)
    result_1 = approx_kernel_based(
        x,
        w,
        num_random_features_X=1,
        num_random_features_Y=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    )
    np.random.seed(0)
    result_2 = approx_kernel_based(
        x,
        w,
        num_random_features_X=1,
        num_random_features_Y=1,
        bootstrap_num_samples=5,
        bootstrap_num_runs=10,
        p_value_adjust_func=np.mean,
    )

    assert result_1 == result_2


@flaky(max_runs=3)
def test_given_weak_dependency_when_perform_kernel_based_test_then_returns_expected_result():
    X = np.random.choice(2, (10000, 100))  # Require a lot of data here for the bootstraps.
    Y = np.sum(X * np.random.uniform(-1, 5), axis=1)

    assert kernel_based(X[:, 0], Y) <= 0.05
    assert kernel_based(np.random.choice(2, (10000, 1)), Y) > 0.05


def test_given_constant_data_when_perform_kernel_based_test_then_returns_expected_result():
    assert kernel_based(np.zeros(100), np.random.normal(0, 1, 100)) == 1.0
    assert kernel_based(np.zeros(100), np.random.normal(0, 1, 100), np.random.normal(0, 1, 100)) == 1.0
    assert not np.isnan(kernel_based(np.random.normal(0, 1, 100), np.random.normal(0, 1, 100), np.zeros(100)))


def test_given_constant_data_when_perform_approx_kernel_based_test_then_returns_expected_result():
    assert approx_kernel_based(np.zeros(100), np.random.normal(0, 1, 100)) == 1.0
    assert approx_kernel_based(np.zeros(100), np.random.normal(0, 1, 100), np.random.normal(0, 1, 100)) == 1.0
    assert not np.isnan(approx_kernel_based(np.random.normal(0, 1, 100), np.random.normal(0, 1, 100), np.zeros(100)))


def _generate_categorical_data():
    x = np.random.normal(0, 1, 1000)
    z = []
    for v in x:
        if v > 0:
            z.append(0)
        else:
            z.append(1)
    y = z + np.random.randn(len(z))
    z = np.array(z).astype(str)
    z[z == "0"] = "Class 1"
    z[z == "1"] = "Class 2"

    return x, y, z


@pytest.fixture
def _preserve_random_generator_state():
    numpy_state = np.random.get_state()
    random_state = random.getstate()
    yield
    np.random.set_state(numpy_state)
    random.setstate(random_state)



import numpy as np
import pandas as pd
from sklearn.neural_network import MLPRegressor

from dowhy.utils.dgp import DataGeneratingProcess


class RandomNeuralNetwork(DataGeneratingProcess):

    TRAINING_SAMPLE_SIZE = 10
    RANDOM_STATE = 0
    DEFAULT_ARCH = (50, 50, 50)
    DEFAULT_ARCH_DICT = {
        "confounder=>treatment": DEFAULT_ARCH,
        "confounder=>outcome": DEFAULT_ARCH,
        "effect_modifier=>outcome": DEFAULT_ARCH,
        "treatment=>outcome": DEFAULT_ARCH,
    }
    NAME = "Random Neural Network"

    def __init__(self, **kwargs):
        """
        Understanding Architectures in MLP Regressor
        https://stackoverflow.com/questions/35363530/python-scikit-learn-mlpclassifier-hidden-layer-sizes

        More Information about Random State
        https://stackoverflow.com/questions/42191717/python-random-state-in-splitting-dataset/42197534
        https://stats.stackexchange.com/questions/80407/am-i-creating-bias-by-using-the-same-random-seed-over-and-over
        """
        super().__init__(**kwargs)
        self.arch = kwargs.pop("arch", RandomNeuralNetwork.DEFAULT_ARCH_DICT)
        self.random_state = kwargs.pop("random_state", RandomNeuralNetwork.RANDOM_STATE)
        self.nn = {}
        self.nn["confounder=>treatment"] = MLPRegressor(
            random_state=self.random_state, hidden_layer_sizes=self.arch["confounder=>treatment"]
        )
        self.nn["confounder=>outcome"] = MLPRegressor(
            random_state=self.random_state, hidden_layer_sizes=self.arch["confounder=>outcome"]
        )
        self.nn["effect_modifier=>outcome"] = MLPRegressor(
            random_state=self.random_state, hidden_layer_sizes=self.arch["effect_modifier=>outcome"]
        )
        self.nn["treatment=>outcome"] = MLPRegressor(
            random_state=self.random_state, hidden_layer_sizes=self.arch["treatment=>outcome"]
        )

    def generate_data(self, sample_size):
        self.generation_process()
        confounder = np.random.randn(sample_size, len(self.confounder))
        effect_modifier = np.random.randn(sample_size, len(self.effect_modifier))
        control_value = np.zeros((sample_size, len(self.treatment)))
        treatment_value = np.ones((sample_size, len(self.treatment)))

        treatment = self.nn["confounder=>treatment"].predict(confounder)
        treatment = treatment[:, np.newaxis]
        if self.treatment_is_binary:
            treatment = self.convert_to_binary(treatment)

        if treatment.ndim == 1:
            treatment = np.reshape(treatment, (-1, 1))

        outcome = (
            self.nn["confounder=>outcome"].predict(np.hstack((confounder, effect_modifier)))
            + self.nn["effect_modifier=>outcome"].predict(effect_modifier)
            + self.nn["treatment=>outcome"].predict(np.hstack((confounder, effect_modifier, treatment)))
        )
        y_control = (
            self.nn["confounder=>outcome"].predict(np.hstack((confounder, effect_modifier)))
            + self.nn["effect_modifier=>outcome"].predict(effect_modifier)
            + self.nn["treatment=>outcome"].predict(np.hstack((confounder, effect_modifier, control_value)))
        )
        y_treatment = (
            self.nn["confounder=>outcome"].predict(np.hstack((confounder, effect_modifier)))
            + self.nn["effect_modifier=>outcome"].predict(effect_modifier)
            + self.nn["treatment=>outcome"].predict(np.hstack((confounder, effect_modifier, treatment_value)))
        )

        # Understanding Neural Network weights
        # Refer to this link:https://stackoverflow.com/questions/50937628/mlp-classifier-neurons-weights
        self.weights = {key: self.nn[key].coefs_ for key in self.nn.keys()}
        self.bias = {key: self.nn[key].intercepts_ for key in self.nn.keys()}

        if outcome.ndim == 1:
            outcome = np.reshape(outcome, (-1, 1))
        if y_control.ndim == 1:
            y_control = np.reshape(y_control, (-1, 1))
        if y_treatment.ndim == 1:
            y_treatment = np.reshape(y_treatment, (-1, 1))
        self.true_value = np.mean(y_treatment - y_control, axis=0)

        return pd.DataFrame(
            np.hstack((effect_modifier, confounder, treatment, outcome)),
            columns=self.effect_modifier + self.confounder + self.treatment + self.outcome,
        )

    def generation_process(self):
        X = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.confounder))
        y = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.treatment))
        self.nn["confounder=>treatment"].fit(X, y)

        X = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.confounder) + len(self.effect_modifier))
        y = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.outcome))
        self.nn["confounder=>outcome"].fit(X, y)

        X = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.effect_modifier))
        y = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.outcome))
        self.nn["effect_modifier=>outcome"].fit(X, y)

        X = np.random.randn(
            RandomNeuralNetwork.TRAINING_SAMPLE_SIZE,
            len(self.confounder) + len(self.effect_modifier) + len(self.treatment),
        )
        y = np.random.randn(RandomNeuralNetwork.TRAINING_SAMPLE_SIZE, len(self.outcome))
        self.nn["treatment=>outcome"].fit(X, y)

    def __str__(self):
        rep = super().__str__()

        header = """
        Random Neural Network Data Generating Process
        ---------------------------------------------
        """
        rep += """
        arch:{}
        nn:{}
        random_state:{}
        """.format(
            self.arch, self.nn, self.random_state
        )

        rep = header + rep

        return rep



from typing import Any, List, Optional, Union

import pandas as pd

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.propensity_score_estimator import PropensityScoreEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class PropensityScoreStratificationEstimator(PropensityScoreEstimator):
    """Estimate effect of treatment by stratifying the data into bins with
    identical common causes.

    Straightforward application of the back-door criterion.

    Supports additional parameters as listed below.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        num_strata: Union[str, int] = "auto",
        clipping_threshold: int = 10,
        propensity_score_model: Optional[Any] = None,
        propensity_score_column: str = "propensity_score",
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param num_strata: Number of bins by which data will be stratified.
            Default is automatically determined.
        :param clipping_threshold: Mininum number of treated or control units
            per strata. Default=10
        :param propensity_score_model: The model used to compute propensity
            score. Can be any classification model that supports fit() and
            predict_proba() methods. If None, use
            LogisticRegression model as the default.
        :param propensity_score_column: Column name that stores the propensity
        score. Default='propensity_score'
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            propensity_score_model=propensity_score_model,
            propensity_score_column=propensity_score_column,
            num_strata=num_strata,
            clipping_threshold=clipping_threshold,
            **kwargs,
        )

        self.logger.info("Using Propensity Score Stratification Estimator")

        # setting method-specific parameters
        self.num_strata = num_strata
        self.clipping_threshold = clipping_threshold

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        super().fit(data, treatment_name, outcome_name, effect_modifier_names=effect_modifier_names)
        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        return self

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        if data is None:
            data = self._data
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value
        if self.propensity_score_column not in data:
            self.estimate_propensity_score_column(data)

        clipped = None
        # Infer the right strata based on clipping threshold
        if self.num_strata == "auto":
            # 0.5 because there are two values for the treatment
            clipping_t = self.clipping_threshold
            num_strata = 0.5 * data.shape[0] / clipping_t
            # To be conservative and allow most strata to be included in the
            # analysis
            strata_found = False
            while not strata_found:
                self.logger.info("'num_strata' selected as {}".format(num_strata))
                try:
                    clipped = self._get_strata(num_strata, self.clipping_threshold)
                    num_ret_strata = clipped.groupby(["strata"]).count().reset_index()
                    # At least 90% of the strata should be included in analysis
                    if num_ret_strata.shape[0] >= 0.5 * num_strata:
                        strata_found = True
                    else:
                        num_strata = int(num_strata / 2)
                        self.logger.info(
                            f"Less than half the strata have at least {self.clipping_threshold} data points. Selecting fewer number of strata."
                        )
                        if num_strata < 2:
                            raise ValueError(
                                "Not enough data to generate at least two strata. This error may be due to a high value of 'clipping_threshold'."
                            )
                except ValueError:
                    self.logger.info(
                        "No strata found with at least {} data points. Selecting fewer number of strata".format(
                            self.clipping_threshold
                        )
                    )
                    num_strata = int(num_strata / 2)
                    if num_strata < 2:
                        raise ValueError(
                            "Not enough data to generate at least two strata. This error may be due to a high value of 'clipping_threshold'."
                        )
        else:
            clipped = self._get_strata(self.num_strata, self.clipping_threshold)

        # sum weighted outcomes over all strata  (weight by treated population)
        weighted_outcomes = clipped.groupby("strata").agg(
            {self._treatment_name[0]: ["sum"], "dbar": ["sum"], "d_y": ["sum"], "dbar_y": ["sum"]}
        )
        weighted_outcomes.columns = ["_".join(x) for x in weighted_outcomes.columns.to_numpy().ravel()]
        treatment_sum_name = self._treatment_name[0] + "_sum"
        control_sum_name = "dbar_sum"

        weighted_outcomes["d_y_mean"] = weighted_outcomes["d_y_sum"] / weighted_outcomes[treatment_sum_name]
        weighted_outcomes["dbar_y_mean"] = weighted_outcomes["dbar_y_sum"] / weighted_outcomes["dbar_sum"]
        weighted_outcomes["effect"] = weighted_outcomes["d_y_mean"] - weighted_outcomes["dbar_y_mean"]
        total_treatment_population = weighted_outcomes[treatment_sum_name].sum()
        total_control_population = weighted_outcomes[control_sum_name].sum()
        total_population = total_treatment_population + total_control_population
        self.logger.debug(
            "Total number of data points is {0}, including {1} from treatment and {2} from control.".format(
                total_population, total_treatment_population, total_control_population
            )
        )

        if target_units == "att":
            est = (
                weighted_outcomes["effect"] * weighted_outcomes[treatment_sum_name]
            ).sum() / total_treatment_population
        elif target_units == "atc":
            est = (weighted_outcomes["effect"] * weighted_outcomes[control_sum_name]).sum() / total_control_population
        elif target_units == "ate":
            est = (
                weighted_outcomes["effect"]
                * (weighted_outcomes[control_sum_name] + weighted_outcomes[treatment_sum_name])
            ).sum() / total_population
        else:
            raise ValueError("Target units string value not supported")

        # TODO - how can we add additional information into the returned estimate?
        #        such as how much clipping was done, or per-strata info for debugging?
        estimate = CausalEstimate(
            estimate=est,
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
            propensity_scores=data[self.propensity_score_column],
        )

        estimate.add_estimator(self)
        return estimate

    def _get_strata(self, num_strata, clipping_threshold):
        # sort the dataframe by propensity score
        # create a column 'strata' for each element that marks what strata it belongs to
        num_rows = self._data[self._outcome_name].shape[0]
        self._data["strata"] = (
            (self._data[self.propensity_score_column].rank(ascending=True) / num_rows) * num_strata
        ).round(0)
        # for each strata, count how many treated and control units there are
        # throw away strata that have insufficient treatment or control

        self._data["dbar"] = 1 - self._data[self._treatment_name[0]]  # 1-Treatment
        self._data["d_y"] = self._data[self._treatment_name[0]] * self._data[self._outcome_name]
        self._data["dbar_y"] = self._data["dbar"] * self._data[self._outcome_name]
        stratified = self._data.groupby("strata")
        clipped = stratified.filter(
            lambda strata: min(
                strata.loc[strata[self._treatment_name[0]] == 1].shape[0],
                strata.loc[strata[self._treatment_name[0]] == 0].shape[0],
            )
            > clipping_threshold
        )
        self.logger.debug(
            "After using clipping_threshold={0}, here are the number of data points in each strata:\n {1}".format(
                clipping_threshold, clipped.groupby(["strata", self._treatment_name[0]])[self._outcome_name].count()
            )
        )
        if clipped.empty:
            raise ValueError(
                "Method requires strata with number of data points per treatment > clipping_threshold (={0}). No such strata exists. Consider decreasing 'num_strata' or 'clipping_threshold' parameters.".format(
                    clipping_threshold
                )
            )
        return clipped

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ",".join(estimand.outcome_variable) + "~"
        # TODO -- fix: we are actually conditioning on positive treatment (d=1)
        var_list = estimand.treatment_variable + estimand.get_backdoor_variables()
        expr += "+".join(var_list)
        return expr



from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class DistanceMatchingEstimator(CausalEstimator):
    """Simple matching estimator for binary treatments based on a distance
    metric.

    Supports additional parameters as listed below.

    """

    # allowed types of distance metric
    Valid_Dist_Metric_Params = ["p", "V", "VI", "w"]

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        num_matches_per_unit: int = 1,
        distance_metric: str = "minkowski",
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param num_matches_per_unit: The number of matches per data point.
            Default=1.
        :param distance_metric: Distance metric to use. Default="minkowski"
            that corresponds to Euclidean distance metric with p=2.
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            num_matches_per_unit=num_matches_per_unit,
            distance_metric=distance_metric,
            **kwargs,
        )

        self.num_matches_per_unit = num_matches_per_unit
        self.distance_metric = distance_metric

        # Dictionary of any user-provided params for the distance metric
        # that will be passed to sklearn nearestneighbors
        self.distance_metric_params = {}
        for param_name in self.Valid_Dist_Metric_Params:
            param_val = getattr(self, param_name, None)
            if param_val is not None:
                self.distance_metric_params[param_name] = param_val

        self.logger.info("INFO: Using Distance Matching Estimator")

        self.matched_indices_att = None
        self.matched_indices_atc = None

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        exact_match_cols=None,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param exact_match_cols: List of column names whose values should be
        exactly matched. Typically used for columns with discrete values.
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self.exact_match_cols = exact_match_cols

        self._set_effect_modifiers(effect_modifier_names)

        # Check if the treatment is one-dimensional
        if len(self._treatment_name) > 1:
            error_msg = str(self.__class__) + "cannot handle more than one treatment variable"
            raise Exception(error_msg)
        # Checking if the treatment is binary
        if not pd.api.types.is_bool_dtype(self._data[self._treatment_name[0]]):
            error_msg = "Distance Matching method is applicable only for binary treatments"
            self.logger.error(error_msg)
            raise Exception(error_msg)

        self.logger.debug("Back-door variables used:" + ",".join(self._target_estimand.get_backdoor_variables()))

        self._observed_common_causes_names = self._target_estimand.get_backdoor_variables()
        if self._observed_common_causes_names:
            if self.exact_match_cols is not None:
                self._observed_common_causes_names = [
                    v for v in self._observed_common_causes_names if v not in self.exact_match_cols
                ]
            self._observed_common_causes = self._data[self._observed_common_causes_names]
            # Convert the categorical variables into dummy/indicator variables
            # Basically, this gives a one hot encoding for each category
            # The first category is taken to be the base line.
            self._observed_common_causes = pd.get_dummies(self._observed_common_causes, drop_first=True)
        else:
            self._observed_common_causes = None
            error_msg = "No common causes/confounders present. Distance matching methods are not applicable"
            self.logger.error(error_msg)
            raise Exception(error_msg)

        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        return self

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        if data is None:
            data = self._data
        # this assumes a binary treatment regime
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value
        updated_df = pd.concat(
            [self._observed_common_causes, data[[self._outcome_name, self._treatment_name[0]]]], axis=1
        )
        if self.exact_match_cols is not None:
            updated_df = pd.concat([updated_df, data[self.exact_match_cols]], axis=1)
        treated = updated_df.loc[data[self._treatment_name[0]] == 1]
        control = updated_df.loc[data[self._treatment_name[0]] == 0]
        numtreatedunits = treated.shape[0]
        numcontrolunits = control.shape[0]

        fit_att, fit_atc = False, False
        est = None
        # TODO remove neighbors that are more than a given radius apart
        if target_units == "att":
            fit_att = True
        elif target_units == "atc":
            fit_atc = True
        elif target_units == "ate":
            fit_att = True
            fit_atc = True
        else:
            raise ValueError("Target units string value not supported")

        if fit_att:
            # estimate ATT on treated by summing over difference between matched neighbors
            if self.exact_match_cols is None:
                control_neighbors = NearestNeighbors(
                    n_neighbors=self.num_matches_per_unit,
                    metric=self.distance_metric,
                    algorithm="ball_tree",
                    **self.distance_metric_params,
                ).fit(control[self._observed_common_causes.columns].values)
                distances, indices = control_neighbors.kneighbors(treated[self._observed_common_causes.columns].values)
                self.logger.debug("distances:")
                self.logger.debug(distances)

                att = 0

                for i in range(numtreatedunits):
                    treated_outcome = treated.iloc[i][self._outcome_name].item()
                    control_outcome = np.mean(control.iloc[indices[i]][self._outcome_name].values)
                    att += treated_outcome - control_outcome

                att /= numtreatedunits
                if target_units == "att":
                    est = att
                elif target_units == "ate":
                    est = att * numtreatedunits

                # Return indices in the original dataframe
                self.matched_indices_att = {}
                treated_df_index = treated.index.tolist()
                for i in range(numtreatedunits):
                    self.matched_indices_att[treated_df_index[i]] = control.iloc[indices[i]].index.tolist()
            else:
                grouped = updated_df.groupby(self.exact_match_cols)
                att = 0
                for name, group in grouped:
                    treated = group.loc[group[self._treatment_name[0]] == 1]
                    control = group.loc[group[self._treatment_name[0]] == 0]
                    if treated.shape[0] == 0:
                        continue
                    control_neighbors = NearestNeighbors(
                        n_neighbors=self.num_matches_per_unit,
                        metric=self.distance_metric,
                        algorithm="ball_tree",
                        **self.distance_metric_params,
                    ).fit(control[self._observed_common_causes.columns].values)
                    distances, indices = control_neighbors.kneighbors(
                        treated[self._observed_common_causes.columns].values
                    )
                    self.logger.debug("distances:")
                    self.logger.debug(distances)

                    for i in range(numtreatedunits):
                        treated_outcome = treated.iloc[i][self._outcome_name].item()
                        control_outcome = np.mean(control.iloc[indices[i]][self._outcome_name].values)
                        att += treated_outcome - control_outcome
                        # self.matched_indices_att[treated_df_index[i]] = control.iloc[indices[i]].index.tolist()

                att /= numtreatedunits

                if target_units == "att":
                    est = att
                elif target_units == "ate":
                    est = att * numtreatedunits

        if fit_atc:
            # Now computing ATC
            treated_neighbors = NearestNeighbors(
                n_neighbors=self.num_matches_per_unit,
                metric=self.distance_metric,
                algorithm="ball_tree",
                **self.distance_metric_params,
            ).fit(treated[self._observed_common_causes.columns].values)
            distances, indices = treated_neighbors.kneighbors(control[self._observed_common_causes.columns].values)

            atc = 0
            for i in range(numcontrolunits):
                control_outcome = control.iloc[i][self._outcome_name].item()
                treated_outcome = np.mean(treated.iloc[indices[i]][self._outcome_name].values)
                atc += treated_outcome - control_outcome

            atc /= numcontrolunits

            if target_units == "atc":
                est = atc
            elif target_units == "ate":
                est += atc * numcontrolunits
                est /= numtreatedunits + numcontrolunits

            # Return indices in the original dataframe
            self.matched_indices_atc = {}
            control_df_index = control.index.tolist()
            for i in range(numcontrolunits):
                self.matched_indices_atc[control_df_index[i]] = treated.iloc[indices[i]].index.tolist()

        estimate = CausalEstimate(
            estimate=est,
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
        )

        estimate.add_estimator(self)
        return estimate

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ", ".join(estimand.outcome_variable) + "~"
        var_list = estimand.treatment_variable + estimand.get_backdoor_variables()
        expr += "+".join(var_list)
        return expr



from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
from autogluon import tabular
from autogluon.tabular.configs.hyperparameter_configs import get_hyperparameter_config
from packaging import version

from dowhy.gcm.fcms import ClassificationModel, PredictionModel
from dowhy.gcm.util.general import shape_into_2d


class _AutoGluonModel(PredictionModel):
    def __init__(
        self,
        verbosity: int = 0,
        eval_metric: Optional[str] = None,
        path: Optional[str] = None,
        model_persistence_max_memory: Optional[float] = None,
        **fit_parameters,
    ) -> None:
        self._verbosity = verbosity
        self._fit_parameters = fit_parameters
        self._eval_metric = eval_metric
        self._model_persistence_max_memory = model_persistence_max_memory

        self._feature_names = None
        self._target_column = ["Y"]

        self._auto_gluon_model = tabular.TabularPredictor(
            label=self._target_column[0], verbosity=self._verbosity, eval_metric=self._eval_metric, path=path
        )

        if "hyperparameters" not in self._fit_parameters:
            self._fit_parameters["hyperparameters"] = get_hyperparameter_config("light")
            self._fit_parameters["hyperparameters"].update({"LR": {}})

        if "infer_limit" not in self._fit_parameters:
            self._fit_parameters["infer_limit"] = 0.005

        if version.parse(tabular.version.__version__) >= version.parse("0.4"):
            if "presets" not in self._fit_parameters:
                self._fit_parameters["presets"] = ["optimize_for_deployment"]

    def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
        X, Y = shape_into_2d(X, Y)

        if Y.shape[1] > 1:
            raise RuntimeError("%s supports currently only one dimensional target variables!")

        self._feature_names = ["X" + str(i) for i in range(X.shape[1])]
        self._auto_gluon_model.fit(
            pd.concat(
                [pd.DataFrame(X, columns=self._feature_names), pd.DataFrame(Y, columns=self._target_column)], axis=1
            ),
            **self._fit_parameters,
        )

        if self._model_persistence_max_memory is not None:
            self._auto_gluon_model.persist_models(max_memory=self._model_persistence_max_memory)
        else:
            self._auto_gluon_model.persist_models()

    def predict(self, X: np.ndarray) -> np.ndarray:
        X = shape_into_2d(X)
        return shape_into_2d(self._auto_gluon_model.predict(pd.DataFrame(X, columns=self._feature_names)).to_numpy())

    @property
    def auto_gluon_model(self) -> tabular.TabularPredictor:
        return self._auto_gluon_model

    @property
    def fit_parameters(self) -> Dict[str, Any]:
        return self._fit_parameters


class AutoGluonRegressor(_AutoGluonModel):
    def __init__(self, **auto_gluon_parameters) -> None:
        super().__init__(**auto_gluon_parameters)

    def clone(self):
        return AutoGluonRegressor(verbosity=self._verbosity, eval_metric=self._eval_metric, **self.fit_parameters)


class AutoGluonClassifier(_AutoGluonModel, ClassificationModel):
    def __init__(self, **auto_gluon_parameters) -> None:
        super().__init__(**auto_gluon_parameters)

    def predict_probabilities(self, X: np.ndarray) -> np.ndarray:
        if self.auto_gluon_model is None:
            raise RuntimeError("AutoGluon model has not been fitted!")

        X = shape_into_2d(X)
        result = self.auto_gluon_model.predict_proba(pd.DataFrame(X, columns=self._feature_names))

        return shape_into_2d(result.to_numpy())

    @property
    def classes(self) -> List[str]:
        return self.auto_gluon_model.class_labels

    def clone(self):
        return AutoGluonClassifier(verbosity=self._verbosity, eval_metric=self._eval_metric, **self.fit_parameters)



"""This module provides functions to estimate causal influences.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""
import logging
import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numpy.matlib import repmat

import dowhy.gcm.auto as auto
from dowhy.gcm._noise import compute_data_from_noise, noise_samples_of_ancestors
from dowhy.gcm.cms import ProbabilisticCausalModel, StructuralCausalModel
from dowhy.gcm.constant import EPS
from dowhy.gcm.divergence import estimate_kl_divergence_of_probabilities
from dowhy.gcm.fcms import ClassificationModel, ClassifierFCM, PredictionModel, ProbabilityEstimatorModel
from dowhy.gcm.fitting_sampling import draw_samples
from dowhy.gcm.graph import (
    ConditionalStochasticModel,
    get_ordered_predecessors,
    is_root_node,
    node_connected_subgraph_view,
    validate_causal_dag,
    validate_node,
)
from dowhy.gcm.shapley import ShapleyConfig, estimate_shapley_values
from dowhy.gcm.stats import marginal_expectation
from dowhy.gcm.uncertainty import estimate_entropy_of_probabilities, estimate_variance
from dowhy.gcm.util.general import is_categorical, set_random_seed, shape_into_2d

_logger = logging.getLogger(__name__)


def arrow_strength(
    causal_model: ProbabilisticCausalModel,
    target_node: Any,
    parent_samples: Optional[pd.DataFrame] = None,
    num_samples_conditional: int = 1000,
    max_num_runs: int = 5000,
    tolerance: float = 10**-4,
    n_jobs: int = 1,
    difference_estimation_func: Optional[Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]]] = None,
) -> Dict[Tuple[Any, Any], float]:
    """Computes the causal strength of each edge directed to the target node.
    The strength of an edge is quantified in terms of distance between conditional distributions of the target node in
    the original graph and the imputed graph wherein the edge has been removed and the target node is fed a random
    permutation of the observations of the source node. For more scientific details behind this API, please refer to
    the research paper below.

    **Research Paper**:
    Dominik Janzing, David Balduzzi, Moritz Grosse-Wentrup, Bernhard Schölkopf. *Quantifying Causal Influences*. The
    Annals of Statistics, Vol. 41, No. 5, 2324-2358, 2013.

    :param causal_model: The probabilistic causal model for whose target node we compute the strength of incoming
                         edges for.
    :param target_node: The target node whose incoming edges' strength is to be computed.
    :param parent_samples: Optional samples from the parents of the target_node. If None are given, they are generated
                           based on the provided causal model. Providing observational data can help to mitigate
                           misspecifications in the graph, such as missing interactions between root nodes or
                           confounders.
    :param num_samples_conditional: Sample size to use for estimating the distance between distributions.
    :param max_num_runs: The maximum number of times to resample and estimate the strength to report the average
                         strength.
    :param tolerance: The difference in average strength between two successive runs to terminate early without
                      running it max_num_runs times.
    :param n_jobs: The number of jobs to run in parallel. Set it to -1 to use all processors.
    :param difference_estimation_func: Optional: How to measure the distance between two distributions. By default,
                                       the difference of the variance is estimated for a continuous target node
                                       and the KL divergence for a categorical target node.
    :return: Causal strength of each edge.
    """
    if target_node not in causal_model.graph.nodes:
        raise ValueError("Target node %s can not be found in given graph!" % target_node)
    if is_root_node(causal_model.graph, target_node):
        raise ValueError("Target node %s is a root node, but it requires to have ancestors!" % target_node)

    # Creating a smaller subgraph, which only contains upstream nodes that are connected to the target node.
    sub_causal_model = ProbabilisticCausalModel(node_connected_subgraph_view(causal_model.graph, target_node))
    validate_node(sub_causal_model.graph, target_node)

    ordered_predecessors = get_ordered_predecessors(sub_causal_model.graph, target_node)

    if parent_samples is None:
        parent_samples = draw_samples(sub_causal_model, num_samples_conditional * 10)[ordered_predecessors]

    direct_influences = arrow_strength_of_model(
        sub_causal_model.causal_mechanism(target_node),
        parent_samples[ordered_predecessors].to_numpy(),
        num_samples_from_conditional=num_samples_conditional,
        max_num_runs=max_num_runs,
        tolerance=tolerance,
        n_jobs=n_jobs,
        difference_estimation_func=difference_estimation_func,
    )
    return {(predecessor, target_node): direct_influences[i] for i, predecessor in enumerate(ordered_predecessors)}


def arrow_strength_of_model(
    conditional_stochastic_model: ConditionalStochasticModel,
    input_samples: np.ndarray,
    num_samples_from_conditional: int = 1000,
    max_num_runs: int = 5000,
    tolerance: float = 10**-4,
    n_jobs: int = 1,
    difference_estimation_func: Optional[Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]]] = None,
    input_subsets: Optional[List[List[int]]] = None,
) -> np.ndarray:
    input_samples = shape_into_2d(input_samples)

    if input_subsets is None:
        input_subsets = [[i] for i in range(input_samples.shape[1])]

    if difference_estimation_func is None:
        if isinstance(conditional_stochastic_model, ProbabilityEstimatorModel):
            difference_estimation_func = estimate_kl_divergence_of_probabilities
        else:

            def difference_estimation_func(old, new):
                return np.var(new) - np.var(old)

    if isinstance(conditional_stochastic_model, ProbabilityEstimatorModel):
        samples_creation_method = conditional_stochastic_model.estimate_probabilities
    else:
        samples_creation_method = conditional_stochastic_model.draw_samples

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")

        def parallel_job(subset: List[int], parallel_random_seed: int):
            set_random_seed(parallel_random_seed)

            return _estimate_direct_strength(
                samples_creation_method,
                input_samples,
                subset,
                difference_estimation_func,
                num_samples_from_conditional,
                max_num_runs,
                tolerance,
            )

        random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(input_subsets))
        results = Parallel(n_jobs=n_jobs)(
            delayed(parallel_job)(subset, random_seed) for subset, random_seed in zip(input_subsets, random_seeds)
        )

    if np.any(results == np.inf):
        _logger.warning(
            "At least one arrow strength is infinite. This typically happens if the causal models are "
            "deterministic, i.e. there is no noise or it is extremely small."
        )

    return np.array(results)


def _estimate_direct_strength(
    draw_samples_func: Callable[[np.ndarray], np.ndarray],
    distribution_samples: np.ndarray,
    parents_subset: List[int],
    difference_estimation_func: Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]],
    num_samples_conditional: int,
    max_num_runs: int,
    tolerance: float,
) -> float:
    distribution_samples = shape_into_2d(distribution_samples)

    aggregated_conditional_difference_result = 0
    average_difference_result = 0
    converged_run = 0
    for run, sample in enumerate(distribution_samples):
        tmp_samples = repmat(sample, num_samples_conditional, 1)

        rnd_permutation = np.random.choice(distribution_samples.shape[0], num_samples_conditional, replace=False)

        # Sampling from the conditional distribution based on the current sample.
        conditional_distribution_samples = draw_samples_func(tmp_samples)

        # Sampling from the conditional based on the current sample, but randomizing the inputs of all variables that
        # are in the given subset. By this, we can simulate the impact on the conditional distribution when removing
        # only the incoming edges of the variables in the subset.
        tmp_samples[:, parents_subset] = distribution_samples[:, parents_subset][rnd_permutation]
        cond_dist_removed_arr_samples = draw_samples_func(tmp_samples)

        old_average_difference_result = average_difference_result

        aggregated_conditional_difference_result += difference_estimation_func(
            conditional_distribution_samples, cond_dist_removed_arr_samples
        )

        average_difference_result = aggregated_conditional_difference_result / (run + 1)

        if run >= max_num_runs:
            break
        elif run > 0:
            if old_average_difference_result == 0:
                old_average_difference_result = EPS

            if abs(1 - average_difference_result / old_average_difference_result) < tolerance:
                converged_run += 1
                if converged_run >= 3:
                    break
            else:
                converged_run = 0

    return average_difference_result


def intrinsic_causal_influence(
    causal_model: StructuralCausalModel,
    target_node: Any,
    prediction_model: Union[PredictionModel, ClassificationModel, str] = "approx",
    attribution_func: Optional[Callable[[np.ndarray, np.ndarray], float]] = None,
    num_training_samples: int = 100000,
    num_samples_randomization: int = 7500,
    num_samples_baseline: int = 1000,
    max_batch_size: int = 100,
    auto_assign_quality: auto.AssignmentQuality = auto.AssignmentQuality.GOOD,
    shapley_config: Optional[ShapleyConfig] = None,
) -> Dict[Any, float]:
    """Computes the causal contribution of each upstream noise term of the target node (including the noise of the
    target itself) to the statistical property (e.g. mean, variance) of the target. We call this contribution
    *intrinsic* as noise terms, by definition, do not inherit properties of observed parents. The contribution of each
    noise term is then the *intrinsic* causal contribution of the corresponding node. For more scientific details,
    please refer to the paper below.

    **Research Paper**:
    Janzing et al. *Quantifying causal contributions via structure preserving interventions*. arXiv:2007.00714, 2021.

    :param causal_model: The structural causal model for whose target node we compute the intrinsic causal influence
                         of its ancestors.
    :param target_node: Target node whose statistical property is to be attributed.
    :param prediction_model: Prediction model for estimating the functional relationship between subsets of ancestor
                             noise terms and the target node. This can be an instance of a PredictionModel, the string
                             'approx' or the string 'exact'. With 'exact', the underlying causal models in the graph
                             are utilized directly by propagating given noise inputs through the graph. This is
                             generally more accurate but slow. With 'approx', an appropriate model is selected and
                             trained based on sampled data from the graph, which is less accurate but faster. A more
                             detailed treatment on why we need this parameter is also provided in :ref:`icc`.
    :param attribution_func: Optional attribution function to measure the statistical property of the target node. This
                             function expects two inputs; predictions after the randomization of certain features (i.e.
                             samples from noise nodes) and a baseline where no features were randomized. The baseline
                             predictions can be typically ignored if one is interested in uncertainty measures such as
                             entropy or variance, but they might be relevant if, for instance, these shall be estimated
                             based on the residuals. By default, entropy is used if prediction model is a classifier,
                             variance otherwise.
    :param num_training_samples: Number of samples drawn from the graphical causal model that are used for fitting the
                                 prediction_model (if necessary).
    :param num_samples_randomization: Number of noise samples drawn from the graphical causal model that are used for
                                      evaluating the set function. Here, these samples are samples from
                                      the noise distributions used for randomizing features that are not in the subset.
    :param num_samples_baseline: Number of noise samples drawn from the graphical causal model that are used for
                                 evaluating the set function. Here, these samples are used as fixed observations for
                                 features that are in the subset.
    :param max_batch_size: Maximum batch size for estimating the predictions from evaluation samples. This has a
                           significant impact on the overall memory usage. If set to -1, all samples are used in one
                           batch.
    :param auto_assign_quality: Auto assign quality for the 'approx' prediction_model option.
    :param shapley_config: :class:`~dowhy.gcm.shapley.ShapleyConfig` for the Shapley estimator.
    :return: Intrinsic causal contribution of each ancestor node to the statistical property defined by the
             attribution_func of the target node.
    """
    validate_causal_dag(causal_model.graph)

    # Creating a smaller subgraph, which only contains upstream nodes that are connected to the target node.
    sub_causal_model = StructuralCausalModel(node_connected_subgraph_view(causal_model.graph, target_node))

    data_samples, noise_samples = noise_samples_of_ancestors(sub_causal_model, target_node, num_training_samples)
    node_names = noise_samples.columns
    noise_samples, target_samples = shape_into_2d(noise_samples.to_numpy(), data_samples[target_node].to_numpy())

    target_is_categorical = is_categorical(data_samples[target_node].to_numpy())

    if prediction_model == "approx":
        prediction_model = auto.select_model(noise_samples, target_samples, auto_assign_quality)
        prediction_model.fit(noise_samples, target_samples)

        if target_is_categorical:
            prediction_method = prediction_model.predict_probabilities
        else:
            prediction_method = prediction_model.predict
    elif prediction_model == "exact":

        def exact_model(X: np.ndarray) -> np.ndarray:
            return compute_data_from_noise(sub_causal_model, pd.DataFrame(X, columns=[x for x in node_names]))[
                target_node
            ].to_numpy()

        if target_is_categorical:
            list_of_classes = cast(
                ClassifierFCM, sub_causal_model.causal_mechanism(target_node)
            ).classifier_model.classes

            def prediction_method(X):
                return (shape_into_2d(exact_model(X)) == list_of_classes).astype(float)

        else:
            prediction_method = exact_model
    elif isinstance(prediction_model, str):
        raise ValueError(
            "Invalid value for prediction_model: %s! This should either be an instance of a PredictionModel or"
            "one of the two string options 'exact' or 'approx'." % prediction_model
        )
    else:
        prediction_model.fit(noise_samples, target_samples)
        prediction_method = prediction_model.predict

    if attribution_func is None:
        if target_is_categorical:

            def attribution_func(x, _):
                return -estimate_entropy_of_probabilities(x)

        else:

            def attribution_func(x, _):
                return estimate_variance(x)

    _, noise_samples = noise_samples_of_ancestors(
        sub_causal_model, target_node, num_samples_randomization + num_samples_baseline
    )
    noise_samples = shape_into_2d(noise_samples.to_numpy())

    iccs = _estimate_iccs(
        attribution_func,
        prediction_method,
        noise_samples[:num_samples_randomization],
        noise_samples[num_samples_randomization : num_samples_randomization + num_samples_baseline],
        max_batch_size,
        ShapleyConfig() if shapley_config is None else shapley_config,
    )

    return {node: iccs[i] for i, node in enumerate(node_names)}


def _estimate_iccs(
    attribution_func: Callable[[np.ndarray, np.ndarray], float],
    prediction_method: Callable[[np.ndarray], np.ndarray],
    noise_samples: np.ndarray,
    baseline_noise_samples: np.ndarray,
    max_batch_size: int,
    shapley_config: ShapleyConfig,
):
    target_values = shape_into_2d(prediction_method(baseline_noise_samples))

    def icc_set_function(subset: np.ndarray) -> Union[np.ndarray, float]:
        if np.all(subset == 1):
            # In case of the full subset (no randomization), we get the same predictions as when we apply the
            # prediction method to the samples of interest, since all noise samples are replaced with a sample of
            # interest.
            predictions = target_values
        elif np.all(subset == 0):
            # In case of the empty subset (all are jointly randomize), it boils down to taking the average over all
            # predictions, seeing that the randomization yields the same values for each sample of interest (none of the
            # samples of interest are used to replace a (jointly) 'randomized' sample).
            predictions = repmat(np.mean(prediction_method(noise_samples), axis=0), baseline_noise_samples.shape[0], 1)
        else:
            predictions = marginal_expectation(
                prediction_method,
                feature_samples=noise_samples,
                baseline_samples=baseline_noise_samples,
                baseline_feature_indices=np.arange(0, noise_samples.shape[1])[subset == 1],
                return_averaged_results=True,
                feature_perturbation="randomize_columns_jointly",
                max_batch_size=max_batch_size,
            )
        return attribution_func(shape_into_2d(predictions), target_values)

    return estimate_shapley_values(icc_set_function, noise_samples.shape[1], shapley_config)



"""This module provides functionality to estimate confidence intervals via bootstrapping.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

from typing import Any, Callable, Dict, List, Tuple, Union

import numpy as np
from joblib import Parallel, delayed
from scipy.optimize import minimize
from tqdm import tqdm

from dowhy.gcm import config
from dowhy.gcm.util.general import set_random_seed, shape_into_2d


def estimate_geometric_median(X: np.ndarray) -> np.ndarray:
    def distance_function(x_input: np.ndarray) -> np.ndarray:
        return np.sum(np.sqrt(np.sum((x_input - X) ** 2, axis=1)))

    return minimize(distance_function, np.sum(X, axis=0) / X.shape[0]).x


def confidence_intervals(
    estimation_func: Union[Callable[[], np.ndarray], Callable[[], Dict[Any, float]]],
    confidence_level: float = 0.95,
    num_bootstrap_resamples: int = 20,
    bootstrap_results_summary_func: Callable[[np.ndarray], np.ndarray] = estimate_geometric_median,
    n_jobs: int = 1,
) -> Tuple[Union[np.ndarray, Dict[Any, np.ndarray]], Union[np.ndarray, Dict[Any, np.ndarray]]]:
    """
    Estimates confidence intervals based on the outputs generated by calling the given estimation_func. Since one result
    for each repetition is produced, all results can be summarized by the method defined in
    summary_method_of_bootstrap_results.
    For instance, summary_method_of_bootstrap_results = lambda x: numpy.mean(x, axis=0) to get the mean over all runs.
    By default, the geometric median is returned.

    Currently, the confidence intervals are empirically estimated based on the n-th estimated quantiles (without bias
    correction) of the results, where the quantiles are determined by the given confidence_level.

    NOTE: The outputs of estimation_func are assumed to be pairwise independent. For multidimensional outputs of
    estimation_func, this could be violated and should be kept in mind. For instance, when evaluating the outcome of
    interventions in a graph like X -> Y -> Z, the confidence intervals are estimate independently for X, Y and Z
    although they have a strong dependency. If estimation_func returns one dimensional results, as for instance when
    estimating the direct arrow strength, then there should be no problem.

    **Example usage with numpy array output:**

        >>> def estimation_func() -> np.ndarray:
        >>>     return direct_arrow_strength_of_model(causal_model, parent_data)
        >>>
        >>> arrow_strengths, confidence_intervals = confidence_intervals(estimation_func)

    **Example usage with dictionary output:**

        >>> def estimation_func() -> Dict[Any, float]:
        >>>     return distribution_change(
        >>>             causal_dag, original_observations, outlier_observations, 'X3')
        >>>
        >>> mean_contributions, confidence_intervals = confidence_intervals(estimation_func)

    More details about the estimation of confidence intervals via bootstrapping can be found `here <https://ocw.mit.edu/courses/mathematics/18-05-introduction-to-probability-and-statistics-spring-2014/readings/MIT18_05S14_Reading24.pdf>`_.

    :param estimation_func: Function that generates a non-deterministic output for which the confidence interval(s) are
           estimated.
    :param confidence_level: Confidence level of the interval.
    :param num_bootstrap_resamples: Number of samples generated by estimation_func, i.e. number of times
           is called. The higher the number, the more accurate the results and intervals,
           but the slower the runtime.
    :param bootstrap_results_summary_func: Function that takes a numpy array with all results as an input and returns
           a single (potentially multidimensional) value/vector. For instance, the mean or median
           over all results.
    :param n_jobs: Number of parallel jobs. Each repetition can be estimated in parallel.
           However, since many other functions of the library are already running in parallel (
           such as distribution change), this is set to 1 by default. Only if it is certain that
           the estimation_func is not running in parallel internally (e.g. when performing
           interventions), this should be set to a different value.
    :return: A tuple (summarized result over all repetitions based on
             summary_method_of_bootstrap_results, confidence interval for each dimension/variable)
    """
    if num_bootstrap_resamples < 1:
        raise ValueError("Number of repetitions should be greater than 0, but got %d" % num_bootstrap_resamples)

    def estimation_func_with_random_seed(random_seed: int) -> Union[np.ndarray, Dict[Any, float]]:
        set_random_seed(random_seed)
        return estimation_func()

    random_seeds = np.random.randint(np.iinfo(np.int32).max, size=num_bootstrap_resamples)
    all_results = Parallel(n_jobs=n_jobs)(
        delayed(estimation_func_with_random_seed)(random_seed)
        for i, random_seed in enumerate(
            tqdm(
                random_seeds,
                position=0,
                leave=True,
                disable=not config.show_progress_bars,
                desc="Estimating bootstrap interval...",
            )
        )
    )

    if isinstance(all_results[0], dict):
        all_results: List[Dict[Any, float]]
        tmp_dict = {}
        for result in all_results:
            for key, value in result.items():
                tmp_dict.setdefault(key, []).append(value)

        summary_result = bootstrap_results_summary_func(np.column_stack([v for v in tmp_dict.values()]))

        return {key: summary_result[i] for i, key in enumerate(tmp_dict)}, {
            key: _estimate_percentile_bounds(np.array(tmp_dict[key]).squeeze(), confidence_level) for key in tmp_dict
        }
    else:
        all_results: np.ndarray = shape_into_2d(np.array(all_results))

        return bootstrap_results_summary_func(all_results), np.array(
            [_estimate_percentile_bounds(all_results[:, i], confidence_level) for i in range(all_results.shape[1])]
        )


def _estimate_percentile_bounds(X: np.ndarray, quantile: float) -> np.ndarray:
    if X.ndim > 1:
        raise ValueError("Estimate bounds currently only supports one dimensional inputs!")

    return np.array([np.percentile(X, (1 - quantile) * 100), np.percentile(X, quantile * 100)])



import networkx as nx
import numpy as np
import pandas as pd
from flaky import flaky
from sklearn.ensemble import HistGradientBoostingClassifier, HistGradientBoostingRegressor
from sklearn.linear_model import ElasticNetCV, LassoCV, LinearRegression, LogisticRegression, RidgeCV
from sklearn.naive_bayes import GaussianNB
from sklearn.pipeline import Pipeline

from dowhy.gcm import ProbabilisticCausalModel
from dowhy.gcm.auto import AssignmentQuality, assign_causal_mechanisms
from dowhy.gcm.ml import AutoGluonClassifier, AutoGluonRegressor


def _generate_linear_regression_data():
    X = np.random.normal(0, 1, (1000, 5))
    Y = np.sum(X * np.random.uniform(-5, 5, X.shape[1]), axis=1)

    return X, Y


def _generate_non_linear_regression_data():
    X = np.random.normal(0, 1, (1000, 5))
    Y = np.sum(np.log(abs(X)), axis=1)

    return X, Y


def _generate_linear_classification_data():
    X = np.random.normal(0, 1, (1000, 5))
    Y = (np.sum(X * np.random.uniform(-5, 5, X.shape[1]), axis=1) > 0).astype(str)

    return X, Y


def _generate_non_classification_data():
    X = np.random.normal(0, 1, (1000, 5))
    Y = (np.sum(np.exp(X), axis=1) > np.median(np.sum(np.exp(X), axis=1))).astype(str)

    return X, Y


@flaky(max_runs=3)
def test_given_linear_regression_problem_when_auto_assign_causal_models_with_good_quality_returns_linear_model():
    X, Y = _generate_linear_regression_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.GOOD)
    assert isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, LinearRegression)


@flaky(max_runs=3)
def test_given_linear_regression_problem_when_auto_assign_causal_models_with_better_quality_returns_linear_model():
    X, Y = _generate_linear_regression_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.BETTER)
    assert isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, LinearRegression)


@flaky(max_runs=3)
def test_given_non_linear_regression_problem_when_auto_assign_causal_models_with_good_quality_returns_non_linear_model():
    X, Y = _generate_non_linear_regression_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.GOOD)
    assert isinstance(
        causal_model.causal_mechanism("Y").prediction_model.sklearn_model, HistGradientBoostingRegressor
    ) or isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, Pipeline)


@flaky(max_runs=3)
def test_given_non_linear_regression_problem_when_auto_assign_causal_models_with_better_quality_returns_non_linear_model():
    X, Y = _generate_non_linear_regression_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.BETTER)
    assert not isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, LinearRegression)
    assert not isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, LassoCV)
    assert not isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, ElasticNetCV)
    assert not isinstance(causal_model.causal_mechanism("Y").prediction_model.sklearn_model, RidgeCV)


@flaky(max_runs=3)
def test_given_linear_classification_problem_when_auto_assign_causal_models_with_good_quality_returns_linear_model():
    X, Y = _generate_linear_classification_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.GOOD)
    assert isinstance(causal_model.causal_mechanism("Y").classifier_model.sklearn_model, LogisticRegression)


@flaky(max_runs=3)
def test_given_linear_classification_problem_when_auto_assign_causal_models_with_better_quality_returns_linear_model():
    X, Y = _generate_linear_classification_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.BETTER)
    assert isinstance(causal_model.causal_mechanism("Y").classifier_model.sklearn_model, LogisticRegression)


@flaky(max_runs=3)
def test_given_non_linear_classification_problem_when_auto_assign_causal_models_with_good_quality_returns_non_linear_model():
    X, Y = _generate_non_classification_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.GOOD)
    assert isinstance(
        causal_model.causal_mechanism("Y").classifier_model.sklearn_model, HistGradientBoostingClassifier
    ) or isinstance(causal_model.causal_mechanism("Y").classifier_model.sklearn_model, Pipeline)


@flaky(max_runs=3)
def test_given_non_linear_classification_problem_when_auto_assign_causal_models_with_better_quality_returns_non_linear_model():
    X, Y = _generate_non_classification_data()

    causal_model = ProbabilisticCausalModel(
        nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")])
    )
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.BETTER)
    assert not isinstance(causal_model.causal_mechanism("Y").classifier_model.sklearn_model, LogisticRegression)
    assert not isinstance(causal_model.causal_mechanism("Y").classifier_model.sklearn_model, GaussianNB)


@flaky(max_runs=3)
def test_given_polynomial_regression_data_with_categorical_input_when_auto_assign_causal_models_then_does_not_raise_error():
    X = np.column_stack(
        [np.random.choice(2, 100, replace=True).astype(str), np.random.normal(0, 1, (100, 2)).astype(object)]
    ).astype(object)
    Y = []
    for i in range(X.shape[0]):
        Y.append(X[i, 1] * X[i, 2] if X[i, 0] == "0" else X[i, 1] + X[i, 2])

    Y = np.array(Y)

    causal_model = ProbabilisticCausalModel(nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y")]))
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.GOOD)
    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.BETTER, override_models=True)


@flaky(max_runs=3)
def test_given_polynomial_classification_data_with_categorical_input_when_auto_assign_causal_models_then_does_not_raise_error():
    X = np.random.normal(0, 1, (100, 2))
    Y = []

    for x in X:
        if x[0] * x[1] > 0:
            Y.append("Class 0")
        else:
            Y.append("Class 1")

    Y = np.array(Y)

    causal_model = ProbabilisticCausalModel(nx.DiGraph([("X0", "Y"), ("X1", "Y")]))
    data = {"X" + str(i): X[:, i] for i in range(X.shape[1])}
    data.update({"Y": Y})

    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.BETTER)
    assign_causal_mechanisms(causal_model, pd.DataFrame(data), quality=AssignmentQuality.GOOD, override_models=True)


def test_when_auto_called_from_main_namespace_returns_no_attribute_error():
    from dowhy import gcm

    _ = gcm.auto.AssignmentQuality.GOOD


def test_when_using_best_quality_then_returns_auto_gluon_model():
    causal_model = ProbabilisticCausalModel(nx.DiGraph([("X", "Y")]))

    assign_causal_mechanisms(causal_model, pd.DataFrame({"X": [1], "Y": [1]}), quality=AssignmentQuality.BEST)
    assert isinstance(causal_model.causal_mechanism("Y").prediction_model, AutoGluonRegressor)

    assign_causal_mechanisms(
        causal_model, pd.DataFrame({"X": [1], "Y": ["Class 1"]}), quality=AssignmentQuality.BEST, override_models=True
    )
    assert isinstance(causal_model.causal_mechanism("Y").classifier_model, AutoGluonClassifier)



import networkx as nx
import numpy as np
import pandas as pd
from flaky import flaky

from dowhy.gcm import (
    InvertibleStructuralCausalModel,
    RejectionResult,
    auto,
    fit,
    kernel_based,
    refute_causal_structure,
    refute_invertible_model,
)


def _generate_simple_non_linear_data() -> pd.DataFrame:
    X = np.random.normal(loc=0, scale=1, size=5000)
    Y = X**2 + np.random.normal(loc=0, scale=1, size=5000)
    Z = np.exp(-Y) + np.random.normal(loc=0, scale=1, size=5000)

    return pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))


@flaky(max_runs=5)
def test_refute_causal_structure_collider():
    # collider: X->Z<-Y
    collider_dag = nx.DiGraph([("X", "Z"), ("Y", "Z")])
    X = np.random.normal(size=500)
    Y = np.random.normal(size=500)
    Z = 2 * X + 3 * Y + np.random.normal(size=500)
    data = pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))
    rejection_result, rejection_summary = refute_causal_structure(collider_dag, data)

    assert rejection_result == RejectionResult.NOT_REJECTED
    assert rejection_summary["X"]["local_markov_test"] == dict()
    assert rejection_summary["X"]["edge_dependence_test"] == dict()
    assert rejection_summary["Y"]["local_markov_test"] == dict()
    assert rejection_summary["Y"]["edge_dependence_test"] == dict()
    assert rejection_summary["Z"]["local_markov_test"] == dict()
    assert rejection_summary["Z"]["edge_dependence_test"]["X"]["success"] == True
    assert rejection_summary["Z"]["edge_dependence_test"]["Y"]["success"] == True


@flaky(max_runs=5)
def test_refute_causal_structure_chain():
    # chain: X->Z->Y
    chain_dag = nx.DiGraph([("X", "Z"), ("Z", "Y")])
    X = np.random.normal(size=500)
    Z = 2 * X + np.random.normal(size=500)
    Y = 3 * Z + np.random.normal(size=500)
    data = pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))
    rejection_result, rejection_summary = refute_causal_structure(chain_dag, data)

    assert rejection_result == RejectionResult.NOT_REJECTED
    assert rejection_summary["X"]["local_markov_test"] == dict()
    assert rejection_summary["X"]["edge_dependence_test"] == dict()
    assert rejection_summary["Z"]["local_markov_test"] == dict()
    assert rejection_summary["Z"]["edge_dependence_test"]["X"]["success"] == True
    assert rejection_summary["Y"]["local_markov_test"]["success"] == True
    assert rejection_summary["Y"]["edge_dependence_test"]["Z"]["success"] == True


@flaky(max_runs=5)
def test_refute_causal_structure_fork():
    # fork: X<-Z->Y
    fork_dag = nx.DiGraph([("Z", "X"), ("Z", "Y")])
    Z = np.random.normal(size=500)
    X = 2 * Z + np.random.normal(size=500)
    Y = 3 * Z + np.random.normal(size=500)
    data = pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))
    rejection_result, rejection_summary = refute_causal_structure(fork_dag, data)

    assert rejection_result == RejectionResult.NOT_REJECTED
    assert rejection_summary["Z"]["local_markov_test"] == dict()
    assert rejection_summary["Z"]["edge_dependence_test"] == dict()
    assert rejection_summary["X"]["local_markov_test"]["success"] == True
    assert rejection_summary["X"]["edge_dependence_test"]["Z"]["success"] == True
    assert rejection_summary["Y"]["local_markov_test"]["success"] == True
    assert rejection_summary["Y"]["edge_dependence_test"]["Z"]["success"] == True


@flaky(max_runs=5)
def test_refute_causal_structure_general():
    # general DAG: X<-Z->Y, X->Y
    general_dag = nx.DiGraph([("Z", "X"), ("Z", "Y"), ("X", "Y")])
    Z = np.random.normal(size=500)
    X = 2 * Z + np.random.normal(size=500)
    Y = 2 * Z + 3 * X + np.random.normal(size=500)
    data = pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))
    rejection_result, rejection_summary = refute_causal_structure(general_dag, data)

    assert rejection_result == RejectionResult.NOT_REJECTED
    assert rejection_summary["Z"]["local_markov_test"] == dict()
    assert rejection_summary["Z"]["edge_dependence_test"] == dict()
    assert rejection_summary["X"]["local_markov_test"] == dict()
    assert rejection_summary["X"]["edge_dependence_test"]["Z"]["success"] == True
    assert rejection_summary["Y"]["local_markov_test"] == dict()
    assert rejection_summary["Y"]["edge_dependence_test"]["Z"]["success"] == True
    assert rejection_summary["Y"]["edge_dependence_test"]["X"]["success"] == True


@flaky(max_runs=5)
def test_refute_causal_structure_adjusted_p_values():
    # fork: X<-Z->Y
    fork_dag = nx.DiGraph([("Z", "X"), ("Z", "Y")])
    Z = np.random.normal(size=500)
    X = 2 * Z + np.random.normal(size=500)
    Y = 3 * Z + np.random.normal(size=500)
    data = pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))
    rejection_result, rejection_summary = refute_causal_structure(fork_dag, data, fdr_control_method="fdr_bh")

    assert (
        rejection_summary["X"]["local_markov_test"]["fdr_adjusted_p_value"]
        >= rejection_summary["X"]["local_markov_test"]["p_value"]
    )
    assert (
        rejection_summary["X"]["edge_dependence_test"]["Z"]["fdr_adjusted_p_value"]
        >= rejection_summary["X"]["edge_dependence_test"]["Z"]["p_value"]
    )
    assert (
        rejection_summary["Y"]["local_markov_test"]["fdr_adjusted_p_value"]
        >= rejection_summary["Y"]["local_markov_test"]["p_value"]
    )
    assert (
        rejection_summary["Y"]["edge_dependence_test"]["Z"]["fdr_adjusted_p_value"]
        >= rejection_summary["Y"]["edge_dependence_test"]["Z"]["p_value"]
    )


def test_when_using_refute_causal_structure_without_fdrc_then_nans_for_adjusted_p_values_are_returned():
    fork_dag = nx.DiGraph([("Z", "X"), ("Z", "Y")])
    Z = np.random.normal(size=500)
    X = 2 * Z + np.random.normal(size=500)
    Y = 3 * Z + np.random.normal(size=500)
    data = pd.DataFrame(data=dict(X=X, Y=Y, Z=Z))
    _, rejection_summary = refute_causal_structure(fork_dag, data, fdr_control_method=None)

    assert np.isnan(rejection_summary["X"]["local_markov_test"]["fdr_adjusted_p_value"])
    assert np.isnan(rejection_summary["X"]["edge_dependence_test"]["Z"]["fdr_adjusted_p_value"])
    assert np.isnan(rejection_summary["Y"]["local_markov_test"]["fdr_adjusted_p_value"])
    assert np.isnan(rejection_summary["Y"]["edge_dependence_test"]["Z"]["fdr_adjusted_p_value"])

    assert not np.isnan(rejection_summary["X"]["local_markov_test"]["p_value"])
    assert not np.isnan(rejection_summary["X"]["edge_dependence_test"]["Z"]["p_value"])
    assert not np.isnan(rejection_summary["Y"]["local_markov_test"]["p_value"])
    assert not np.isnan(rejection_summary["Y"]["edge_dependence_test"]["Z"]["p_value"])


@flaky(max_runs=2)
def test_given_non_linear_data_and_correct_dag_when_refute_invertible_model_then_not_reject_model():
    data = _generate_simple_non_linear_data()

    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("X", "Y"), ("Y", "Z")]))  # X->Y->Z
    auto.assign_causal_mechanisms(causal_model, data, auto.AssignmentQuality.GOOD)

    fit(causal_model, data)

    assert (
        refute_invertible_model(
            causal_model, data, independence_test=lambda x, y: kernel_based(x, y, bootstrap_num_runs=5)
        )
        == RejectionResult.NOT_REJECTED
    )
    assert (
        refute_invertible_model(
            causal_model,
            data,
            independence_test=lambda x, y: kernel_based(x, y, bootstrap_num_runs=5),
            fdr_control_method="fdr_bh",
        )
        == RejectionResult.NOT_REJECTED
    )


@flaky(max_runs=2)
def test_given_non_linear_data_and_incorrect_dag_when_refute_invertible_model_then_reject_model():
    data = _generate_simple_non_linear_data()

    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("Z", "Y"), ("Y", "X")]))  # X<-Y<-Z
    auto.assign_causal_mechanisms(causal_model, data, auto.AssignmentQuality.GOOD)

    fit(causal_model, data)

    assert (
        refute_invertible_model(
            causal_model, data, independence_test=lambda x, y: kernel_based(x, y, bootstrap_num_runs=5)
        )
        == RejectionResult.REJECTED
    )
    assert (
        refute_invertible_model(
            causal_model,
            data,
            independence_test=lambda x, y: kernel_based(x, y, bootstrap_num_runs=5),
            fdr_control_method="fdr_bh",
        )
        == RejectionResult.REJECTED
    )


@flaky(max_runs=3)
def test_given_non_linear_data_and_incorrect_dag_with_collider_when_refute_invertible_model_then_reject_model():
    data = _generate_simple_non_linear_data()
    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("X", "Y"), ("Z", "Y")]))  # X->Y<-Z
    auto.assign_causal_mechanisms(causal_model, data, auto.AssignmentQuality.GOOD)

    fit(causal_model, data)

    assert (
        refute_invertible_model(
            causal_model, data, independence_test=lambda x, y: kernel_based(x, y, bootstrap_num_runs=10)
        )
        == RejectionResult.REJECTED
    )
    assert (
        refute_invertible_model(
            causal_model,
            data,
            independence_test=lambda x, y: kernel_based(x, y, bootstrap_num_runs=10),
            fdr_control_method="fdr_bh",
        )
        == RejectionResult.REJECTED
    )
